"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import unittest
import numpy
import pandas
import json
import gemmi
import os
import shutil
import sys
import tempfile
import hashlib
from servalcat import utils
from servalcat.__main__ import main

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve

root = os.path.abspath(os.path.dirname(__file__))
data = {}

def download():
    wd = os.path.join(root, "7dy0")
    if not os.path.exists(wd):
        os.mkdir(wd)

    url_md5 = (('half1', 'https://files.wwpdb.org/pub/emdb/structures/EMD-30913/other/emd_30913_half_map_1.map.gz', '0bf81c0057d3017972e7cf53e8a8df77'),
               ('half2', 'https://files.wwpdb.org/pub/emdb/structures/EMD-30913/other/emd_30913_half_map_2.map.gz', '5bba9c32a95d6675fe0387342af5a439'),
               ('mask', 'https://files.wwpdb.org/pub/emdb/structures/EMD-30913/masks/emd_30913_msk_1.map', 'e69403cec5be349b097c8080ce5bb1fd'),
               ('pdb', 'https://files.wwpdb.org/pub/pdb/data/structures/divided/pdb/dy/pdb7dy0.ent.gz', '7ddca7f90c2ebc4b73f2c14136cbda0c'),
               ('mmcif', 'https://files.wwpdb.org/pub/pdb/data/structures/divided/mmCIF/dy/7dy0.cif.gz', 'a0221f46f5e21bb6f53d5f8751d22444'))

    for name, url, md5 in url_md5:
        dst = os.path.join(root, "7dy0", os.path.basename(url))
        if not os.path.exists(dst):
            print("downloading {}".format(url))
            urlretrieve(url, dst)

        with open(dst, "rb") as f:
            print("testing md5sum of {}".format(dst))
            md5f = hashlib.md5(f.read()).hexdigest()
        assert md5 == md5f
        data[name] = dst

download()

def load_halfmaps():
    return [utils.fileio.read_ccp4_map(data[f]) for f in ("half1", "half2")]
def load_mask():
    return utils.fileio.read_ccp4_map(data["mask"])[0]
def load_pdb():
    return utils.fileio.read_structure(data["pdb"])
def load_mmcif():
    return utils.fileio.read_structure(data["mmcif"])

class TestFunctions(unittest.TestCase):
    def test_h_add(self):
        st = load_mmcif()
        self.assertEqual(st[0].count_hydrogen_sites(), 0)
        monlib = utils.restraints.load_monomer_library(st)
        utils.restraints.add_hydrogens(st, monlib, "elec")
        self.assertEqual(st[0].count_hydrogen_sites(), 841)
        bad_h = [numpy.allclose(cra.atom.pos.tolist(), [0,0,0]) for cra in st[0].all()]
        self.assertEqual(sum(bad_h), 0)
    # test_h_add()

    def test_halfmaps_stats(self):
        maps = load_halfmaps()
        mask = load_mask()
        d_min = utils.maps.nyquist_resolution(maps[0][0])
        self.assertAlmostEqual(d_min, 1.6, places=3)
        hkldata = utils.maps.mask_and_fft_maps(maps, d_min, mask)
        hkldata.setup_relion_binning()
        utils.maps.calc_noise_var_from_halfmaps(hkldata)
        numpy.testing.assert_allclose(hkldata.binned_df.var_signal, [3.01280911e+04, 2.73253902e+03, 1.12831493e+03, 7.78811837e+02,
                                                                     8.62027816e+02, 6.75563283e+02, 5.04824473e+02, 3.68568664e+02,
                                                                     3.08512467e+02, 2.24858349e+02, 2.01293139e+02, 2.07807437e+02,
                                                                     1.95582559e+02, 1.88518597e+02, 2.58906187e+02, 3.79117006e+02,
                                                                     4.89705517e+02, 5.22415086e+02, 4.74064846e+02, 3.95723101e+02,
                                                                     3.14365888e+02, 2.47997260e+02, 2.01949742e+02, 1.87939993e+02,
                                                                     1.47301900e+02, 1.22057347e+02, 9.44857640e+01, 7.59628917e+01,
                                                                     6.01736445e+01, 4.40971512e+01, 3.45815865e+01, 2.77571337e+01,
                                                                     2.12604387e+01, 1.78079757e+01, 1.32672151e+01, 1.04391673e+01,
                                                                     8.40114179e+00, 6.95752329e+00, 6.39158281e+00, 6.01332523e+00,
                                                                     4.49567083e+00, 3.62891391e+00, 3.11307538e+00, 2.18187534e+00,
                                                                     1.61615641e+00, 1.75082127e+00, 2.51349128e+00, 2.90669157e+00,
                                                                     3.03402711e+00, 1.70745712e+00, 9.39018836e-01, 1.36677463e+00,
                                                                     6.77851414e-01],
                                      rtol=1e-5)
    # test_halfmaps_stats()
# class TestFunctions

class TestSPACommands(unittest.TestCase):
    def setUp(self):
        self.wd = tempfile.mkdtemp(prefix="servaltest_")
        os.chdir(self.wd)
        print("In", self.wd)
    # setUp()
    
    def tearDown(self):
        os.chdir(root)
        shutil.rmtree(self.wd)
    # tearDown()

    def test_fofc(self):
        sys.argv = ["", "fofc", "--halfmaps", data["half1"], data["half2"],
                    "--model", data["pdb"], "-d", "1.9"]
        main()
        self.assertTrue(os.path.isfile("diffmap.mtz"))
        mtz = gemmi.read_mtz_file("diffmap.mtz")
        self.assertEqual(mtz.nreflections, 208238)
    # test_fofc()

    def test_fsc(self):
        sys.argv = ["", "fsc", "--halfmaps", data["half1"], data["half2"],
                    "--model", data["pdb"], "--mask", data["mask"]]
        main()
        self.assertTrue(os.path.isfile("fsc.dat"))
        df = pandas.read_table("fsc.dat", comment="#", sep=r"\s+")
        
        numpy.testing.assert_allclose(df.fsc_FC_full,
                                      [0.999132, 0.981112, 0.960192, 0.937756, 0.945335, 0.935267,
                                       0.922341, 0.893095, 0.906006, 0.891224, 0.903129, 0.920052,
                                       0.917194, 0.925133, 0.941335, 0.951829, 0.96452, 0.968749,
                                       0.965054, 0.962379, 0.960315, 0.960591, 0.958181, 0.956107,
                                       0.943763, 0.937195, 0.936865, 0.934519, 0.922404, 0.887335,
                                       0.857482, 0.822188, 0.798473, 0.774636, 0.732629, 0.658573,
                                       0.621479, 0.598489, 0.55252, 0.50756, 0.428187, 0.302367,
                                       0.288519, 0.23337, 0.165119, 0.13633, 0.095398, 0.064603,
                                       0.029095, 0.011709, -0.003319, 0.022268, 0.02982],
                                      rtol=1e-4)
        numpy.testing.assert_allclose(df.fsc_half_masked_corrected,
                                      [0.99998, 0.999854, 0.999561, 0.998637, 0.998478, 0.998217, 0.998024, 0.996446, 0.995316,
                                       0.992014, 0.991325, 0.992381, 0.990588, 0.989501, 0.990386, 0.991547, 0.992241, 0.99298,
                                       0.991725, 0.989109, 0.983623, 0.976955, 0.970857, 0.96608, 0.958066, 0.947261, 0.925901,
                                       0.896515, 0.865408, 0.833411, 0.781538, 0.754205, 0.703446, 0.653649, 0.587084, 0.513894,
                                       0.459777, 0.393584, 0.349775, 0.333019, 0.260333, 0.231767, 0.187071, 0.150573, 0.108945,
                                       0.102676, 0.140925, 0.134152, 0.136187, 0.104752, 0.063349, 0.080262, 0.016436],
                                      rtol=1e-4)
    # test_fsc()

    @unittest.skipUnless(utils.refmac.check_version(), "refmac unavailable")
    def test_refine(self):
        sys.argv = ["", "refine_spa", "--halfmaps", data["half1"], data["half2"],
                    "--model", data["pdb"], "--mask_for_fofc", data["mask"],
                    "--trim_fofc_mtz", "--resolution", "1.9", "--ncycle", "5", "--cross_validation"]
        main()
        self.assertTrue(os.path.isfile("refined_fsc.json"))
        self.assertTrue(os.path.isfile("refined.mmcif"))
        self.assertTrue(os.path.isfile("diffmap.mtz"))
        self.assertTrue(os.path.isfile("shifted_refined.log"))
        self.assertTrue(os.path.isfile("refined_expanded.pdb"))

        fscavgs = [float(l.split()[-1]) for l in open("shifted_refined.log") if l.startswith("Average Fourier shell correlation")]
        self.assertEqual(len(fscavgs), 6)
        self.assertGreater(fscavgs[-1], 0.6)
        
        st = utils.fileio.read_structure("refined_expanded.pdb")
        self.assertEqual(len(st[0]), 4)

        sys.argv = ["", "util", "json2csv", "refined_fsc.json"]
        main()
        # TODO check result?
    # test_refine()

    def test_trim(self):
        sys.argv = ["", "trim", "--maps", data["half1"], data["half2"],
                    "--model", data["pdb"], "--mask", data["mask"],
                    "--no_shift", "--noncubic", "--noncentered"]
        main()
        self.assertTrue(os.path.isfile("emd_30913_half_map_1_trimmed.mrc"))
        self.assertTrue(os.path.isfile("emd_30913_half_map_2_trimmed.mrc"))
        self.assertTrue(os.path.isfile("emd_30913_msk_1_trimmed.mrc"))
        self.assertTrue(os.path.isfile("trim_shifts.json"))

        with open("trim_shifts.json") as ifs:
            shifts = json.load(ifs)
        numpy.testing.assert_allclose(shifts["shifts"], [-1.59999327, -4.79997982, -7.99996636], rtol=1e-5)
        self.assertEqual(shifts["new_grid"], [106, 98, 90])
    # test_trim()

    def test_translate(self):
        pass

    def test_localcc(self):
        sys.argv = ["", "localcc", "--halfmaps", data["half1"], data["half2"],
                    "--model", data["pdb"], "--mask", data["mask"], "--kernel", "5"]
        main()
        self.assertTrue(os.path.isfile("ccmap_r5px_half.mrc"))
        self.assertTrue(os.path.isfile("ccmap_r5px_model.mrc"))

        st = utils.fileio.read_structure(data["pdb"])
        halfcc = utils.fileio.read_ccp4_map("ccmap_r5px_half.mrc")[0]
        modelcc = utils.fileio.read_ccp4_map("ccmap_r5px_model.mrc")[0]

        self.assertAlmostEqual(numpy.mean([modelcc.interpolate_value(cra.atom.pos) for cra in st[0].all()]),
                               0.6416836618309301, places=3)
        self.assertAlmostEqual(numpy.mean([halfcc.interpolate_value(cra.atom.pos) for cra in st[0].all()]),
                               0.6619259582976047, places=3)

    def test_commands(self): # util commands
        sys.argv = ["", "util", "symmodel", "--model", data["pdb"], "--map", data["mask"],
                    "--pg", "D2", "--biomt"]
        main()

        sys.argv = ["", "util", "expand", "--model", "pdb7dy0_asu.pdb"]
        main()

        sys.argv = ["", "util", "h_add", data["pdb"]]
        main()

        # TODO merge_models

        sys.argv = ["", "util", "power", "--map", data["mask"], data["half1"], data["half2"]]
        main()

        sys.argv = ["", "util", "fcalc", "--model", data["pdb"], "-d", "1.7", "--auto_box_with_padding=5"]
        main()

        sys.argv = ["", "util", "nemap", "--halfmaps", data["half1"], data["half2"],
                    "--mask", data["mask"], "--trim_mtz", "-d", "1.7"]
        main()

        sys.argv = ["", "util", "blur", "--hklin", "nemap.mtz", "-B", "100"]
        main()
    # test_commands()
# class TestSPACommands

if __name__ == '__main__':
    unittest.main()




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology

This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import scipy.sparse
from servalcat.utils import logger
from servalcat.xtal import sigmaa
from servalcat import utils
from servalcat import ext
b_to_u = utils.model.b_to_u
u_to_b = utils.model.u_to_b
integr = sigmaa.integr

class LL_Xtal:
    def __init__(self, hkldata, centric_and_selections, free, st, monlib, source="xray", mott_bethe=True,
                 use_solvent=False, use_in_est="all", use_in_target="all"):
        assert source in ("electron", "xray", "neutron")
        self.source = source
        self.mott_bethe = False if source != "electron" else mott_bethe
        self.hkldata = hkldata
        self.is_int = "I" in self.hkldata.df
        self.centric_and_selections = centric_and_selections
        self.free = free
        self.st = st
        self.monlib = monlib
        self.d_min = hkldata.d_min_max()[0]
        self.fc_labs = ["FC0"]
        self.use_solvent = use_solvent
        if use_solvent:
            self.fc_labs.append("FCbulk")
            self.hkldata.df["FCbulk"] = 0j
        self.D_labs = ["D{}".format(i) for i in range(len(self.fc_labs))]
        self.k_overall = numpy.ones(len(self.hkldata.df.index))
        self.b_aniso = None
        self.hkldata.df["k_aniso"] = 1.
        self.use_in_est = use_in_est
        self.use_in_target = use_in_target
        self.ll = None
        self.scaling = sigmaa.LsqScale()
        logger.writeln("will use {} reflections for parameter estimation".format(self.use_in_est))
        logger.writeln("will use {} reflections for refinement".format(self.use_in_target))

    def update_ml_params(self):
        self.b_aniso = sigmaa.determine_ml_params(self.hkldata, self.is_int, self.fc_labs, self.D_labs, self.b_aniso,
                                                   self.centric_and_selections, use=self.use_in_est,
                                                  )#D_trans="splus", S_trans="splus")
        self.hkldata.df["k_aniso"] = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        #determine_mlf_params_from_cc(self.hkldata, self.fc_labs, self.D_labs,
        #                             self.centric_and_selections)


    def update_fc(self):
        if self.st.ncs:
            st = self.st.clone()
            st.expand_ncs(gemmi.HowToNameCopiedChain.Dup, merge_dist=0)
        else:
            st = self.st

        self.hkldata.df[self.fc_labs[0]] = utils.model.calc_fc_fft(st, self.d_min - 1e-6,
                                                                   monlib=self.monlib,
                                                                   source=self.source,
                                                                   mott_bethe=self.mott_bethe,
                                                                   miller_array=self.hkldata.miller_array())
        self.hkldata.df["FC"] = self.hkldata.df[self.fc_labs].sum(axis=1)
        
    def overall_scale(self, min_b=0.1):
        fc_list = [self.hkldata.df[self.fc_labs[0]].to_numpy()]
        if self.use_solvent:
            Fmask = sigmaa.calc_Fmask(self.st, self.d_min - 1e-6, self.hkldata.miller_array())
            fc_list.append(Fmask)

        self.scaling.set_data(self.hkldata, fc_list, self.is_int, sigma_cutoff=0)
        self.scaling.scale()
        self.b_aniso = self.scaling.b_aniso
        b = self.scaling.b_iso
        min_b_iso = self.st[0].calculate_b_aniso_range()[0] # actually min of aniso too
        tmp = min_b_iso + b
        if tmp < min_b: # perhaps better only adjust b_iso that went too small, but we need to recalculate Fc
            logger.writeln(" Adjusting overall B to avoid too small value")
            b += min_b - tmp
        logger.writeln(" Applying overall B to model: {:.2f}".format(b))
        utils.model.shift_b(self.st[0], b)
        k_iso = self.hkldata.debye_waller_factors(b_iso=b)
        self.hkldata.df["k_aniso"] = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        if self.use_solvent:
            solvent_scale = self.scaling.get_solvent_scale(self.scaling.k_sol, self.scaling.b_sol,
                                                      1. / self.hkldata.d_spacings().to_numpy()**2)
            self.hkldata.df[self.fc_labs[-1]] = Fmask * solvent_scale
        if self.is_int:
            o_labs = self.hkldata.df.columns.intersection(["I", "SIGI",
                                                           "I(+)","SIGI(+)", "I(-)", "SIGI(-)"])
            self.hkldata.df[o_labs] /= self.scaling.k_overall**2
        else:
            o_labs = self.hkldata.df.columns.intersection(["FP", "SIGFP",
                                                           "F(+)","SIGF(+)", "F(-)", "SIGF(-)"])
            self.hkldata.df[o_labs] /= self.scaling.k_overall

        for lab in self.fc_labs: self.hkldata.df[lab] *= k_iso
        self.hkldata.df["FC"] = self.hkldata.df[self.fc_labs].sum(axis=1)

        # for next cycle
        self.scaling.k_overall = 1.
        self.scaling.b_iso = 0.
    # overall_scale()

    def calc_target(self): # -LL target for MLF or MLI
        ret = 0
        k_aniso = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        f = sigmaa.mli if self.is_int else sigmaa.mlf
        for i_bin, _ in self.hkldata.binned():
            if self.use_in_target == "all":
                idxes = numpy.concatenate([sel[i] for sel in self.centric_and_selections[i_bin] for i in (1,2)])
            else:
                i = 1 if self.use_in_target == "work" else 2
                idxes = numpy.concatenate([sel[i] for sel in self.centric_and_selections[i_bin]])
            ret += f(self.hkldata.df,
                     self.fc_labs,
                     numpy.vstack([self.hkldata.df[lab].to_numpy()[idxes] for lab in self.D_labs]).T,
                     self.hkldata.df.S.to_numpy()[idxes],
                     k_aniso,
                     idxes)
        return ret * 2 # friedel mates
    # calc_target()

    def calc_stats(self, bin_stats=False):
        stats, overall = sigmaa.calc_r_and_cc(self.hkldata, self.centric_and_selections)
        ret = {"summary": overall}
        ret["summary"]["-LL"] = self.calc_target()
        if bin_stats:
            ret["bin_stats"] = stats
        for lab in "R", "CC":
            logger.writeln(" ".join("{} = {:.4f}".format(x, overall[x]) for x in overall if x.startswith(lab)))
        return ret

    def calc_grad(self, atom_pos, refine_xyz, adp_mode, refine_occ, refine_h, specs=None):
        dll_dab = numpy.zeros(len(self.hkldata.df.FC), dtype=numpy.complex128)
        d2ll_dab2 = numpy.empty(len(self.hkldata.df.index))
        d2ll_dab2[:] = numpy.nan
        blur = utils.model.determine_blur_for_dencalc(self.st, self.d_min / 3) # TODO need more work
        logger.writeln("blur for deriv= {:.2f}".format(blur))
        k_ani = self.hkldata.debye_waller_factors(b_cart=self.b_aniso)
        for i_bin, _ in self.hkldata.binned():
            for c, work, test in self.centric_and_selections[i_bin]:
                if self.use_in_target == "all":
                    cidxes = numpy.concatenate([work, test])
                else:
                    cidxes = work if self.use_in_target == "work" else test
                epsilon = self.hkldata.df.epsilon.to_numpy()[cidxes]
                Fcs = numpy.vstack([self.hkldata.df[lab].to_numpy()[cidxes] for lab in self.fc_labs]).T
                Ds = numpy.vstack([self.hkldata.df[lab].to_numpy()[cidxes] for lab in self.D_labs]).T
                S = self.hkldata.df["S"].to_numpy()[cidxes]
                Fc = (Ds * Fcs).sum(axis=1)
                Fc_abs = numpy.abs(Fc)
                expip = numpy.exp(1j * numpy.angle(Fc))
                if self.is_int:
                    Io = self.hkldata.df.I.to_numpy()
                    sigIo = self.hkldata.df.SIGI.to_numpy()
                    to = Io[cidxes] / sigIo[cidxes] - sigIo[cidxes] / (c+1) / k_ani[cidxes]**2 / S / epsilon
                    tf = k_ani[cidxes] * Fc_abs / numpy.sqrt(sigIo[cidxes])
                    sig1 = k_ani[cidxes]**2 * epsilon * S / sigIo[cidxes]
                    k_num = 0.5 if c == 0 else 0. # acentric:0.5, centric: 0.
                    r = ext.integ_J_ratio(k_num, k_num - 0.5, True, to, tf, sig1, c+1,
                                          integr.exp2_threshold, integr.h, integr.N, integr.ewmax)
                    r *= numpy.sqrt(sigIo[cidxes]) / k_ani[cidxes]
                    g = (2-c) * (Fc_abs - r) / epsilon / S  * Ds[:,0]
                    dll_dab[cidxes] = g * expip
                    #d2ll_dab2[cidxes] = (2-c)**2 / S / epsilon * Ds[0]**2 # approximation
                    #d2ll_dab2[cidxes] = ((2-c) / S / epsilon + ((2-c) * r / k_ani[cidxes] / epsilon / S)**2) * Ds[0]**2
                    d2ll_dab2[cidxes] =  g**2
                else:
                    Fo = self.hkldata.df.FP.to_numpy()[cidxes] / k_ani[cidxes]
                    SigFo = self.hkldata.df.SIGFP.to_numpy()[cidxes] / k_ani[cidxes]
                    if c == 0: # acentric
                        Sigma = 2 * SigFo**2 + epsilon * S
                        X = 2 * Fo * Fc_abs / Sigma
                        m = gemmi.bessel_i1_over_i0(X)
                        g = 2 * (Fc_abs - m * Fo) / Sigma * Ds[:,0]  # XXX assuming 0 is atomic structure
                        dll_dab[cidxes] = g * expip
                        d2ll_dab2[cidxes] = (2 / Sigma - (1 - m / X - m**2) * (2 * Fo / Sigma)**2) * Ds[:,0]**2
                    else:
                        Sigma = SigFo**2 + epsilon * S
                        X = Fo * Fc_abs / Sigma
                        #X = X.astype(numpy.float64)
                        m = numpy.tanh(X)
                        g = (Fc_abs - m * Fo) / Sigma * Ds[:,0]
                        dll_dab[cidxes] = g * expip
                        d2ll_dab2[cidxes] = (1. / Sigma - (Fo / (Sigma * numpy.cosh(X)))**2) * Ds[:,0]**2

        if self.mott_bethe:
            dll_dab *= self.hkldata.d_spacings()**2 * gemmi.mott_bethe_const()
            d2ll_dab2 *= gemmi.mott_bethe_const()**2

        # we need V**2/n for gradient.
        dll_dab_den = self.hkldata.fft_map(data=dll_dab * self.hkldata.debye_waller_factors(b_iso=-blur))
        dll_dab_den.array[:] *= self.hkldata.cell.volume**2 / dll_dab_den.point_count
        #asu = dll_dab_den.masked_asu()
        #dll_dab_den.array[:] *= 1 - asu.mask_array # 0 to use
        
        self.ll = ext.LL(self.st, atom_pos, self.mott_bethe, refine_xyz, adp_mode, refine_occ, refine_h)
        self.ll.set_ncs([x.tr for x in self.st.ncs if not x.given])
        if self.source == "neutron":
            self.ll.calc_grad_n92(dll_dab_den, blur)
        else:
            self.ll.calc_grad_it92(dll_dab_den, blur)

        # second derivative
        if self.source == "neutron":
            self.ll.make_fisher_table_diag_direct_n92(1./self.hkldata.d_spacings().to_numpy(),
                                                      d2ll_dab2)
            self.ll.fisher_diag_from_table_n92()
        else:
            self.ll.make_fisher_table_diag_direct_it92(1./self.hkldata.d_spacings().to_numpy(),
                                                       d2ll_dab2)
            self.ll.fisher_diag_from_table_it92()
        #json.dump(dict(b=ll.table_bs, pp1=ll.pp1, bb=ll.bb),
        #          open("ll_fisher.json", "w"), indent=True)
        #a, (b,c) = ll.fisher_for_coo()
        #json.dump(([float(x) for x in a], ([int(x) for x in b], [int(x) for x in c])), open("fisher.json", "w"))
        if specs is not None:
            self.ll.spec_correction(specs)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import model as model_util
import gemmi
b_to_u = model_util.b_to_u

def parse_atom_spec(s, itk):
    # s: list of keywords
    ret = {}
    while itk < len(s):
        if s[itk].lower().startswith(("chai", "segm")):
            ret["chain"] = s[itk+1]
            itk += 2
        elif s[itk].lower().startswith("resi"):
            ret["resi"] = int(s[itk+1])
            itk += 2
        elif s[itk].lower().startswith("ins"):
            ret["icode"] = s[itk+1] if s[itk+1] != "." else " "
            itk += 2
        elif s[itk].lower().startswith(("atom", "atna", "name")):
            if s[itk+1] == "{":
                idx_close = s[itk+1:].index("}") + itk + 1
                ret["names"] = s[itk+2:idx_close]
                itk = idx_close + 1
            else:
                ret["names"] = [s[itk+1]]
                itk += 2
        elif s[itk].lower().startswith("alt"):
            ret["altloc"] = s[itk+1]
            itk += 2
        elif s[itk].lower().startswith("symm"):
            ret["symm"] = s[itk+1][0].lower() == "y"
        else:
            break

    return ret, itk
# parse_atom_spec()

def parse_from_to(s, itk):
    # s: list of keywords
    ret = {}
    if not s[itk].lower().startswith("from"):
        raise RuntimeError("invalid from_to instruction: {}".format(s))

    if s[itk+1] == "*":
        ret["ifirst"] = None # Refmac sets -9999
    else:
        ret["ifirst"] = int(s[itk+1])

    if s[itk+2].lower() != "to":
        ret["chain"] = s[itk+2]
        assert s[itk+3].lower() == "to"
        itk += 4
    else:
        itk += 3

    if s[itk] == "*":
        ret["ilast"] = None # Refmac sets 9999
    else:
        ret["ilast"] = int(s[itk])

    if "chain" not in ret:
        ret["chain"] = s[itk+1]
        itk += 2
    else:
        itk += 2
        
    return ret, itk
# parse_from_to()

def read_exte(s):
    # using the same variable names as used in read_extra_restraints.f
    ret = dict(defaults={})
    if not s: return ret
    defs = ret["defaults"]
    rest_flag_old = rest_flag = False
    if s[0].lower().startswith("exte"):
        if s[1].lower().startswith("gene"): return ret
        elif s[1].lower().startswith("file"): # XXX not supported
            file_ext_now = s[2]
        elif s[1].lower().startswith("syma"): # symall
            # refmac sets "n" by default if "syma" given - but it is not good!
            defs["symall_block"] = "y" if s[2][0].lower() == "y" else "n"
            if len(s) > 3 and s[3].lower().startswith("excl"): # exclude
                defs["exclude_self_block"] = s[4].lower().startswith("self")
        elif s[1].lower().startswith("typa"): # typeall
            #type_default = 2
            defs["type_default"] = max(0, min(2, int(s[2])))
        elif s[1].lower().startswith("alph"): # alphall
            defs["alpha_default"] = float(s[2])
        elif s[1].lower().startswith("verb"): # verbose
            defs["ext_verbose"] = s[2][0].lower() != "n" and not s[2].lower().startswith("off")
            # print("External verbose is on, i.e.") ...
        elif s[1].lower().startswith("weig"): # weight
            itk = 2
            while itk < len(s): # FIXME check out-of-bounds in s[]
                if s[itk].lower().startswith("scal"):
                    itk += 1
                    if itk >= len(s): break
                    try:
                        defs["scale_sigma_dist"] = float(s[itk]) # scale_sigma_loc
                        itk += 1
                        continue
                    except ValueError:
                        pass
                    for k in ("angl", "tors", "chir", "plan", "dist", "inte"):
                        if s[itk].lower().startswith(k):
                            itk += 1
                            if itk >= len(s): break
                            try:
                                defs["scale_sigma_{}".format(k)] = float(s[itk])
                                itk += 1
                                break
                            except ValueError:
                                pass
                elif s[itk].lower().startswith("sgmn"):
                    defs["sigma_min_loc"] = float(s[itk+1])
                    itk += 2
                elif s[itk].lower().startswith("sgmx"):
                    defs["sigma_max_loc"] = float(s[itk+1])
                    itk += 2
                else:
                    raise RuntimeError("Error==> EXTE keyword interpretation: {}".format(" ".join(s)))
        elif s[1].lower().startswith(("miss", "unde")): # undefined
            defs["ignore_undefined"] = s[2].lower().startswith("igno")
        elif s[1].lower().startswith("hydr"): # hydrogen
            defs["ignore_hydrogens"] = s[2].lower().startswith("igno")
        elif s[1].lower().startswith("cut"): # TODO I think this should be parsed outside
            ret["sd_ext_cut"] = float(s[2])  # as this affects everything (not only following blocks)
        elif s[1].lower().startswith("dmax"):
            defs["dist_max_external"] = float(s[2])
        elif s[1].lower().startswith("dmin"):
            defs["dist_min_external"] = float(s[2])
        elif s[1].lower().startswith("use"):
            defs["use_atoms"] = s[2][0].lower()
            if defs["use_atoms"] not in ("a", "m", "h"):
                logger.writeln("invalid exte use keyword: {}".format(s[2]))
                defs["use_atoms"] = "a"
        elif s[1].lower().startswith("conv"):
            if s[2].lower().startswith("pref"):
                defs["prefix_ch"] = s[3] # ???
        elif s[1].lower().startswith(("dist", "plan", "chir", "angl", "inte", "tors")):
            ret["rest_type"] = s[1][:4].lower()
            itk = 2
            iat = 0
            ret["restr"] = {}
            n_expect = dict(plan=0, dist=2, inte=2, angl=3).get(ret["rest_type"], 4)
            ret["restr"]["specs"] = [None for _ in range(n_expect)]
            while itk < len(s):
                if s[itk].lower().startswith(("firs", "seco", "thir", "four", "next", "atre", "atin")):
                    iat = dict(firs=0, seco=1, thir=2, four=3).get(s[itk][:4].lower(), iat+1)
                    atoms, itk = parse_atom_spec(s, itk+1)
                    if ret["rest_type"] == "plan":
                        ret["restr"]["specs"].append(atoms)
                    else:
                        ret["restr"]["specs"][iat] = atoms
                elif s[itk].lower().startswith("type"):
                    try:
                        ret["restr"]["itype_in"] = int(s[itk+1])
                    except ValueError:
                        ret["restr"]["itype_in"] = dict(o=0, f=2).get(s[itk+1][0].lower(), 1)
                    if not (0 <= ret["restr"]["itype_in"] <= 2):
                        logger.writeln("WARNING: wrong type is given. setting to 2.\n=> {}".format(l))
                        ret["restr"]["itype_in"] = 2
                    itk += 2
                elif s[itk].lower().startswith("symm"): # only for distance
                    ret["restr"]["symm_in"] = s[itk+1][0].lower() == "y"
                    itk += 2
                else:
                    d = dict(valu="value", dmin="dmin", dmax="dmax", smin="smin_value", smax="smax_value",
                             sigm="sigma_value", alph="alpha_in", prob="prob_in")
                    k = s[itk][:4].lower()
                    if k in d:
                        ret["restr"][d[k]] = float(s[itk+1])
                        itk += 2
                    else:
                        logger.writeln("unrecognised key: {}\n=> {}".format(s[itk], l))
        elif s[1].lower().startswith("stac"):
            ret["rest_type"] = "stac"
            ret["restr"] = {}
            ret["restr"]["specs"] = [[] for _ in range(2)]
            itk = 2
            #if s[itk].lower().startswith("dist"):
            ip = 0
            while itk < len(s):
                if s[itk].lower().startswith("plan"):
                    ip = int(s[itk+1])
                    itk += 2
                    if ip not in (1, 2):
                        raise RuntimeError("Problem with stacking instructions. Plane number can be 1 or 2.\n=> {}".format(l))
                elif s[itk].lower().startswith(("firs", "next")):
                    atoms, itk = parse_atom_spec(s, itk+1)
                    ret["restr"]["specs"][ip-1] = atoms
                elif s[itk].lower().startswith(("dist", "sddi", "angl", "sdan", "type")):
                    k = dict(dist="dist_id", sddi="dist_sd", angl="angle_id", sdan="angle_sd", type="type_r")[s[itk][:4].lower()]
                    ret["restr"][k] = float(s[itk+1]) if k != "type_r" else int(s[itk+1])
                    itk += 2
                else:
                    logger.writeln("WARNING: unrecognised keyword: {}\n=> {}".format(s[itk], l))
                    itk += 1
        elif s[1].lower().startswith(("harm", "spec")):
            ret["rest_type"] = s[1][:4].lower() # in Refmac, irest_type = 1 if harm else 2
            ret["restr"] = dict(rectype="", toler=0.5, sigma_t=0.5, sigma_u=2.0 * b_to_u, u_val_incl=False)
            itk = 2
            while itk < len(s):
                if s[itk].lower().startswith("auto"):
                    ret["restr"]["rectype"] = "auto"
                    itk += 1
                elif s[itk].lower().startswith("atin"):
                    ret["restr"]["rectype"] = "atom"
                    atoms, itk = parse_atom_spec(s, itk+1)
                    ret["restr"]["specs"] = [atoms]
                elif s[itk].lower().startswith("resi"):
                    ret["restr"]["rectype"] = "resi"
                    fromto, itk = parse_from_to(s, itk+1)
                    ret["restr"]["specs"] = [fromto]
                    if s[itk].lower().startswith("atom"):
                        ret["restr"]["specs"][0]["atom"] = s[itk+1] # called atom_resi in Refmac
                        itk += 2
                elif s[itk].lower().startswith("sigm"):
                    ret["restr"]["sigma_t"] = float(s[itk+1])
                    itk += 2
                elif s[itk].lower().startswith("tole"):
                    ret["restr"]["toler"] = float(s[itk+1])
                    itk += 2
                elif s[itk].lower().startswith(("uval", "bval")):
                    if len(s) > itk+1 and s[itk+1].lower().startswith("incl"):
                        ret["restr"]["u_val_incl"] = True
                        itk += 2
                    else:
                        ret["restr"]["u_val_incl"] = False
                        itk += 1
                elif s[itk].lower().startswith(("sigb", "sigu")):
                    ret["restr"]["sigma_u"] = float(s[itk+1]) * b_to_u
                    itk += 2
                else:
                    logger.writeln("WARNING: unrecognised keyword: {}\n=> {}".format(s[itk], l))
                    itk += 1

        else:
            logger.writeln("WARNING: cannot parse: {}".format(l))
    return ret
# read_exte()

def read_ridge_params(l, r):
    s = l.split()
    assert s[0].lower().startswith("ridg")
    ntok = len(s)
    if s[1].lower().startswith("dist") and ntok > 2:
        if s[2].lower().startswith("with"):
            r.setdefault("groups", []).append({})
            #r["groups"][-1]["sigma"] = sigma_dist_r
            #r["groups"][-1]["dmax"] = dmax_dist_r
            itk = 3
            while itk < ntok:
                if s[itk].lower().startswith("chai"):
                    r["groups"][-1]["chain"] = s[itk+1]
                    itk += 2
                elif s[itk].lower().startswith("resi"):
                    r["groups"][-1]["resi"] = (int(s[itk+1]), int(s[itk+2]))
                    itk += 3
                elif s[itk].lower().startswith("sigm"):
                    v = float(s[itk+1])
                    if v < 0: v = 0.01
                    r["groups"][-1]["sigma"] = v
                    itk += 2
                elif s[itk].lower().startswith("dmax"):
                    v = float(s[itk+1])
                    if v < 0: v = 4.2
                    r["groups"][-1]["dmax"] = v
                    itk += 2
        elif s[2].lower().startswith("incl") and ntok > 3:
            # a: ridge_dist_include_all
            # h: ridge_dist_include_hbond
            # m: ridge_dist_include_main
            v = s[3][0].lower()
            if v in ("a", "h", "m"): r["include"] = v
        elif s[2].lower().startswith("sigm"):
            v = float(s[3])
            r["sigma"] = v if v > 0 else 0.01
        elif s[2].lower().startswith("dmax"):
            v = float(s[3])
            r["dmax"] = v if v > 0 else 4.2
        elif s[2].lower().startswith("inte") and ntok > 3:
            r["interchain"] = s[3][0].lower() == "y"
        elif s[2].lower().startswith("symm") and ntok > 3:
            r["intersym"] = s[3][0].lower() == "y"
        elif s[2].lower().startswith("long") and ntok > 3:
            r["long_range"] = max(0, int(s[3])) # long_range_residue_gap
        elif s[2].lower().startswith("shor") and ntok > 3:
            r["short_range"] = max(0, int(s[3])) # short_range_residue_gap
        elif s[2].lower().startswith("hydr"):
            r["hydrogen"] = ntok < 4 or s[3][0].lower() == "i" # hydrogens_include
        elif s[2].lower().startswith("side") and ntok > 3:
            r["sidechain"] = s[3][0].lower() == "i" # rb_side_chain_include
        elif s[2].lower().startswith("filt"):
            r["bvalue_filter"] = True
            if s[3].lower().startswith("bran"):
                v = float(s[4])
                r["bvalue_filter_range"] = v if v > 0 else 2.0
        else:
            logger.writeln("WARNING: unrecognised keyword: {}\n=> {}".format(s[2], l))
    elif s[1].lower().startswith(("atom", "posi")): # not used
        r["sigma_pos"] = float(s[2]) if ntok > 2 else 0.1
    elif s[1].lower().startswith(("bval", "uval")) and ntok > 2:
        if s[2].lower().startswith("diff"):
            itk = 3
            while itk < ntok:
                if s[itk].lower().startswith("sigm"):
                    if ntok > itk + 1:
                        r["sigma_uval_diff"] = float(s[itk+1])
                        itk += 2
                    else:
                        r["sigma_uval_diff"] = 0.025
                        itk += 1
                elif s[itk].lower().startswith("dmax"): 
                    if ntok > itk + 1:
                        r["dmax_uval_diff"] = float(s[itk+1])
                        itk += 2
                    else:
                        r["dmax_uval_diff"] = 4.2
                        itk += 1
                elif s[itk].lower().startswith("dmwe"): # not used
                    if ntok > itk + 1:
                        r["dmax_uval_weight"] = float(s[itk+1]) * b_to_u
                        itk += 2
                    else:
                        r["dmax_uval_weight"] = 3.0
                        itk += 1
                else:
                    itk += 1
        else:
            r["sigma_b"] = float(s[2])
            r["sigma_u"] = float(s[2])
    else:
        logger.writeln("WARNING: unrecognised keyword: {}\n=> {}".format(s[1], l))

    return r
# read_ridge_params()

def read_occupancy_params(l, r):
    s = l.split()
    if not s[0].lower().startswith("occu"):
        return r
    ntok = len(s)
    r.setdefault("groups", {}) # {igr: [{selection}, {selection}, ..]}
    r.setdefault("const", []) # [[is_comp, group_ids]]
    r.setdefault("ncycle", 0) # 0 means no refine
    if (ntok > 4 and
        s[1].lower().startswith("grou") and
        s[2].lower().startswith("id")):
        igr = s[3]
        gr = r["groups"].setdefault(igr, [])
        gr.append({})
        itk = 4
        while itk < ntok:
            if s[itk].lower().startswith(("chai", "segm")):
                gr[-1]["chains"] = []
                itk += 1
                while itk < ntok and not s[itk].lower().startswith(("resi","atom","alt")):
                    gr[-1]["chains"].append(s[itk])
                    itk += 1
            elif s[itk].lower().startswith("resi"):
                if s[itk+1].lower().startswith("from"):
                    gr[-1]["resi_from"] = gemmi.SeqId(s[itk+2])
                    if s[itk+3].lower().startswith("to"):
                        gr[-1]["resi_to"] = gemmi.SeqId(s[itk+4])
                    itk += 5
                else:
                    gr[-1]["resi"] = gemmi.SeqId(s[itk+1])
                    itk += 2
            elif s[itk].lower().startswith("atom"):
                gr[-1]["atom"] = s[itk+1]
                itk += 2
            elif s[itk].lower().startswith("alt"):
                gr[-1]["alt"] = s[itk+1]
                itk += 2
    elif (ntok > 4 and
          s[1].lower().startswith("grou") and
          s[2].lower().startswith("alts")):
        r["const"].append((s[3].lower().startswith("comp"), s[4:]))
    elif ntok > 1 and s[1].lower().startswith("refi"):
        if ntok > 3 and s[2].lower().startswith("ncyc"):
            r["ncycle"] = max(int(s[3]), r["ncycle"])
        elif r["ncycle"] == 0:
            r["ncycle"] = 1 # default

    return r
# read_occupancy_params()

def read_restr_params(l, r):
    s = l.split()

    def read_tors_params(itk):
        ret = {"flag": True}
        if s[itk].lower().startswith("none"): # remove all
            ret["flag"] = False
            itk += 1
        elif s[itk].lower().startswith("resi"):
            ret["residue"] = s[itk+1]
            itk += 2
        elif s[itk].lower().startswith("grou"):
            ret["group"] = s[itk+1] # group_name_tors_restr_o
            itk += 2
        elif s[itk].lower().startswith("link"):
            ret["link"] = s[itk+1] #link_name_tors_restr_o
            itk += 2
        else:
            pass # raise error?
        while itk < len(s):
            if s[itk].lower().startswith("name"):
                itk += 1
                ret["tors_name"] = s[itk] # RES_NAME_TORS_NAME_O
            elif s[itk].lower().startswith("valu"):
                itk += 1
                ret["tors_value"] = float(s[itk]) # RES_NAME_TORS_VALUE_O
            elif s[itk].lower().startswith("sigm"):
                itk += 1
                ret["tors_sigma"] = float(s[itk]) # RES_NAME_TORS_SIGMA_O
            elif s[itk].lower().startswith("peri"):
                itk += 1
                ret["tors_period"] = int(s[itk]) # RES_NAME_TORS_PERIOD_O
            itk += 1
        return ret, itk
    # read_tors_params()
    
    if not s[0].lower().startswith("rest"):
        return r
    if s[1].lower().startswith("excl"):
        # restr_params.f90 subroutine exclude_restraints
        #r["exclude"] = True
        pass
    elif s[1].lower().startswith("conf"): # not implemented in Refmac
        pass
    elif s[1].lower().startswith(("bp", "pair", "base")):
        if s[2].lower().startswith("dist"):
            r["plane_dist"] = True
        else:
            r["basepair"] = True
            # TODO read dnarna_params.txt
    elif s[1].lower().startswith("tors") and len(s) > 2:
        itk = 2
        if s[itk].lower().startswith("hydr"):
            itk += 1
            if s[itk].lower().startswith("incl"):
                itk += 1
                r["htors_restraint"] = True
                if s[itk].lower().startswith("all"): # this is servalcat default for now
                    r["htors_restraint_type"] = "all"
                elif s[itk].lower().startswith("sele"):
                    r["htors_restraint_type"] = "sele"
                    r["htors_restraint_list"] = s[itk+1:]
            else:
                itk += 1
                r["htors_restraint"] = False
        elif s[itk].lower().startswith("fbas"):
            pass # set user_basepair_file
        elif s[itk].lower().startswith("incl"):
            tmp, itk = read_tors_params(itk+1)
            r.setdefault("torsion_include", []).append(tmp)
        elif s[itk].lower().startswith("excl"):
            # Should warn if value/sigma/period given?
            tmp, itk = read_tors_params(itk+1)
            r.setdefault("torsion_exclude", []).append(tmp)
    elif s[1].lower().startswith("resi"):
        pass
    elif s[1].lower().startswith("chir"):
        pass
# read_restr_params()

def read_make_params(l, r):
    # TODO: hout,ribo,valu,spec,form,sdmi,segi
    s = l.split()
    assert s[0].lower().startswith("make")
    itk = 1
    ntok = len(s)
    # keyword, key, func, possible, default
    keys = (("hydr", "hydr", lambda x: x[0].lower(), set("aynf"), "a"),
            ("hout", "hout", lambda x: x[0].lower() in ("y", "p"), (True, False), True),
            ("chec", "check", lambda x: "0" if x.lower().startswith(("none", "0"))
             else ("n" if x.lower().startswith(("liga", "n"))
                   else ("y" if x.lower().startswith(("all", "y")) else None)),
             set("0ny"), None), # no default
            ("newl", "newligand", lambda x: x.lower().startswith(("c", "y", "noex")), (True, False), False),
            ("buil", "build", lambda x: x[0].lower(), set("ny"), "n"),
            ("pept", "pept", lambda x: x[0].lower(), set("yn"), "y"),
            ("link", "link", lambda x: x[0].lower(), set("ynd0"), "y"),
            ("suga", "sugar", lambda x: x[0].lower(), set("ynds"), "y"),
            #("conn", "conn", lambda x: x[0].lower(), set("nyd0"), "n"), # TODO read conn_tolerance? (make conn tole val)
            ("symm", "symm", lambda x: x[0].lower(), set("ync"), "y"),
            ("chai", "chain", lambda x: x[0].lower(), set("yn"), "y"),
            ("cisp", "cispept", lambda x: x[0].lower(), set("yn"), "y"),
            ("ss", "ss", lambda x: x[0].lower(), set("ydn"), "y"),
            ("exit", "exit", lambda x: x[0].lower() == "y", (True, False), True),
            )
    while itk < ntok:
        found = False
        for k, t, f, p, d in keys:
            if s[itk].lower().startswith(k):
                if itk + 1 < len(s):
                    r[t] = f(s[itk+1])
                    if r[t] not in p:
                        raise SystemExit("Invalid make instruction: {}".format(l))
                    itk += 2
                elif d is not None:
                    r[t] = d # set default
                    itk += 1
                else:
                    raise SystemExit("Invalid make instruction: {}".format(l))
                break
        else: # if no keywords match (can raise an error if all make keywords are implemented)
            itk += 1
    return r
# read_make_params()

def parse_line(l, ret):
    s = l.split()
    ntok = len(s)
    if ntok == 0: return
    if s[0].lower().startswith("exte"):
        ret.setdefault("exte", []).append(read_exte(s))
    elif s[0].lower().startswith("make"):
        read_make_params(l, ret.setdefault("make", {}))
    elif s[0].lower().startswith(("sour", "scat")):
        k = s[1].lower()
        if k.startswith("em"):
            ret["source"] = "em"
        elif k.startswith("e"):
            ret["source"] = "ec"
        elif k.startswith("n"):
            ret["source"] = "ne"
        else:
            ret["source"] = "xr"
        # TODO check mb, lamb
    elif s[0].lower().startswith("refi"): # TODO support this. Note that only valid with hklin
        ret.setdefault("refi", {})
        itk = 1
        while itk < ntok:
            if s[itk].startswith("type"):
                if itk+1 < ntok and s[itk+1].startswith("unre"):
                    ret["refi"]["type"] = "unre"
                itk += 2
            else:
                itk += 1
    elif s[0].lower().startswith("dist"):
        try:
            ret["wbond"] = float(s[1])
        except:
            pass
        # TODO read sdex, excu, dele, dmxe, dmne
    elif s[0].lower().startswith("ridg"):
        read_ridge_params(l, ret.setdefault("ridge", {}))
    elif s[0].lower().startswith("occu"):
        read_occupancy_params(l, ret.setdefault("occu", {}))
    elif s[0].lower().startswith("rest"):
        read_restr_params(l, ret.setdefault("restr", {}))
    elif s[0].lower().startswith("angl") and ntok > 1:
        ret["wangle"] = float(s[1])
    elif s[0].lower().startswith("tors") and ntok > 1:
        ret["wtors"] = float(s[1])
    elif s[0].lower().startswith(("bfac", "temp", "bval")):
        pass # TODO
    elif s[0].lower().startswith("plan") and ntok > 1:
        ret["wplane"] = float(s[1])
    elif s[0].lower().startswith("chir") and ntok > 1:
        ret["wchir"] = float(s[1])
        # TODO read calp
    elif s[0].lower().startswith(("vdwr", "vand", "nonb")) and ntok > 1:
        itk = 1
        try:
            ret["wvdw"] = float(s[itk])
            itk += 1
        except ValueError:
            pass
        # TODO read maxr, over, sigm, incr, chan, vdwc, excl
# parse_line()

def get_lines(lines, depth=0):
    ret = []
    cont = ""
    for l in lines:
        if "!" in l: l = l[:l.index("!")]
        if "#" in l: l = l[:l.index("#")]
        l = l.strip()
        if not l: continue
        if l[0] == "@":
            f = l[1:]
            try:
                yield from get_lines(open(f).readlines(), depth+1)
            except RuntimeError:
                return
            continue
        if l.split()[-1] == "-":
            cont += l[:l.rfind("-")] + " "
            continue
        if cont:
            l = cont + l
            cont = ""
        if l.split()[0].lower().startswith("end"):
            # refmac stops reading keywords when "exit" is seen in stdin or a file
            # but won't do this from nested files
            if depth == 1:
                raise RuntimeError
            break
        yield l
# get_lines()
            
def update_params(ret, inputs):
    if not inputs:
        return
    for l in get_lines(inputs):
        parse_line(l, ret)
# update_keywords()

def parse_keywords(inputs):
    ret = {"make":{}, "ridge":{}, "refi":{}}
    update_params(ret, inputs)
    return ret
# parse_keywords()

if __name__ == "__main__":
    import sys
    import json
    print("waiting for input")
    ret = {} #{"make":{}, "ridge":{}, "refi":{}}
    for l in get_lines(sys.stdin):
        parse_line(l, ret)
    print()
    print("Parsed:")
    print(json.dumps(ret, indent=1))



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from . import logger
from . import symmetry
from . import fileio
from . import hkl
from . import model
from . import maps
from . import refmac
from . import restraints
from . import commands

def make_loggraph_str(df, main_title, title_labs, s2=None, float_format=None):
    if s2 is not None:
        df = df.copy()
        df.insert(0, "1/resol^2", s2)
    ret = "$TABLE: {} :\n".format(main_title)
    ret += "$GRAPHS\n"
    all_labs = list(df.columns)
    for t, labs in title_labs:
        if s2 is not None: labs = ["1/resol^2"] + labs
        ret += ": {} :A:{}:\n".format(t, ",".join(str(all_labs.index(l)+1) for l in labs))
    ret += "$$\n"
    lines = df.to_string(index=False, index_names=False, header=True, float_format=float_format).splitlines()
    ret += lines[0] + "\n$$\n$$\n"
    ret += "\n".join(lines[1:]) + "\n$$\n"
    return ret
# make_loggraph_str()




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
from servalcat.utils import fileio
from servalcat.utils import symmetry
from servalcat.utils import model
from servalcat.utils import hkl
from servalcat.utils import restraints
from servalcat.utils import maps
from servalcat.refmac import refmac_keywords
from servalcat.refine.refine import Geom
from servalcat import ext
import os
import gemmi
import numpy
import scipy.spatial
import pandas
import json
import re
import argparse

def add_arguments(p):
    subparsers = p.add_subparsers(dest="subcommand")

    # show
    parser = subparsers.add_parser("show", description = 'Show file info supported by the program')
    parser.add_argument('files', nargs='+')

    # json2csv
    parser = subparsers.add_parser("json2csv", description = 'Convert json to csv for plotting')
    parser.add_argument('json')
    parser.add_argument('-o', '--output_prefix')

    # symmodel
    parser = subparsers.add_parser("symmodel", description="Add symmetry annotation to model")
    parser.add_argument('--model', required=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--map', help="Take box size from the map")
    group.add_argument('--cell', type=float, nargs=6, metavar=("a", "b", "c", "alpha", "beta", "gamma"),
                       help="Box size")
    sym_group = parser.add_argument_group("symmetry")
    symmetry.add_symmetry_args(sym_group, require_pg=True)
    parser.add_argument('--contacting_only', action="store_true", help="Filter out non-contacting NCS")
    parser.add_argument('--chains', nargs="*", action="append", help="Select chains to keep")
    parser.add_argument('--howtoname', choices=["dup", "short", "number"], default="short",
                        help="How to decide new chain IDs in expanded model (default: short); "
                        "dup: use original chain IDs (with different segment IDs), "
                        "short: use unique new IDs, "
                        "number: add number to original chain ID")
    parser.add_argument('--biomt', action="store_true", help="Add BIOMT also")
    parser.add_argument('-o', '--output_prfix')
    parser.add_argument('--pdb', action="store_true", help="Write a pdb file")
    parser.add_argument('--cif', action="store_true", help="Write a cif file")

    # helical_biomt
    parser = subparsers.add_parser("helical_biomt", description="generate BIOMT of helical reconstruction for PDB deposition")
    parser.add_argument('--model', required=True)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--map', help="Take box size from the map")
    group.add_argument('--cell', type=float, nargs=6, metavar=("a", "b", "c", "alpha", "beta", "gamma"),
                       help="Box size")
    sym_group = parser.add_argument_group("symmetry")
    symmetry.add_symmetry_args(sym_group, require_pg=True)
    parser.add_argument('--start', type=int)
    parser.add_argument('--end', type=int)
    parser.add_argument('--howtoname', choices=["dup", "short", "number"], default="short",
                        help="How to decide new chain IDs in expanded model (default: short); "
                        "dup: use original chain IDs (with different segment IDs), "
                        "short: use unique new IDs, "
                        "number: add number to original chain ID")
    parser.add_argument('-o', '--output_prfix')

    # expand
    parser = subparsers.add_parser("expand", description="Expand symmetry")
    parser.add_argument('--model', required=True)
    parser.add_argument('--chains', nargs="*", action="append", help="Select chains to keep")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--howtoname', choices=["dup", "short", "number"], default="short",
                       help="How to decide new chain IDs in expanded model (default: short); "
                       "dup: use original chain IDs (with different segment IDs), "
                       "short: use unique new IDs, "
                       "number: add number to original chain ID")
    group.add_argument("--split", action="store_true", help="split file for each operator")
    parser.add_argument('-o', '--output_prfix')
    parser.add_argument('--pdb', action="store_true", help="Write a pdb file")
    parser.add_argument('--cif', action="store_true", help="Write a cif file")

    # h_add
    parser = subparsers.add_parser("h_add", description = 'Add hydrogen in riding position')
    parser.add_argument('model')
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('-o','--output')
    parser.add_argument("--pos", choices=["elec", "nucl"], default="elec")

    # add_op3
    parser = subparsers.add_parser("add_op3", description = "Add OP3 atoms to 5' ends")
    parser.add_argument('model')
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('-o','--output')

    # map_peaks
    parser = subparsers.add_parser("map_peaks", description = 'List density peaks and write a coot script')
    parser.add_argument('--model', required=True, help="Model")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--map', help="Map file")
    group.add_argument('--mtz', help="MTZ for map file")
    parser.add_argument('--mtz_labels', default="DELFWT,PHDELWT", help='F,PHI labels (default: %(default)s)')
    parser.add_argument('--oversample_pixel', type=float, help='Desired pixel spacing in map (Angstrom)')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--sigma_level', type=float, help="Threshold map level in sigma unit")
    group.add_argument('--abs_level', type=float, help="Threshold map level in absolute unit")
    parser.add_argument('--blob_pos', choices=["peak", "centroid"], default="centroid",
                       help="default: %(default)s")
    parser.add_argument('--min_volume', type=float, default=0.3, help="minimum blob volume (default: %(default).1f)")
    parser.add_argument('--max_volume', type=float, help="maximum blob volume (default: none)")
    parser.add_argument('-o','--output_prefix', default="peaks")
    
    # h_density
    parser = subparsers.add_parser("h_density", description = 'Hydrogen density analysis')
    parser.add_argument('--model', required=True, help="Model with hydrogen atoms")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--map', help="Fo-Fc map file")
    group.add_argument('--mtz', help="MTZ for Fo-Fc map file")
    parser.add_argument('--mtz_labels', default="DELFWT,PHDELWT", help='F,PHI labels (default: %(default)s)')
    parser.add_argument('--oversample_pixel', type=float, help='Desired pixel spacing in map (Angstrom)')
    #parser.add_argument("--source", choices=["electron", "xray", "neutron"], default="electron")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--sigma_level', type=float, help="Threshold map level in sigma unit")
    group.add_argument('--abs_level', type=float, help="Threshold map level in absolute unit")
    parser.add_argument('--max_dist', type=float, default=0.5, help="max distance between peak and hydrogen position in the model (default: %(default).1f)")
    parser.add_argument('--blob_pos', choices=["peak", "centroid"], default="centroid",
                       help="default: %(default)s")
    parser.add_argument('--min_volume', type=float, default=0.3, help="minimum blob volume (default: %(default).1f)")
    parser.add_argument('--max_volume', type=float, default=3, help="maximum blob volume (default: %(default).1f)")
    parser.add_argument('-o','--output_prefix')

    # fix_link
    parser = subparsers.add_parser("fix_link", description = 'Fix LINKR/_struct_conn records in the model')
    parser.add_argument('model')
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--bond_margin', type=float, default=1.3, help='(default: %(default).1f)')
    parser.add_argument('--metal_margin', type=float, default=1.1, help='(default: %(default).1f)')
    parser.add_argument('-o','--output', help="Default: input_fixlink.{pdb|mmcif}")

    # merge_models
    parser = subparsers.add_parser("merge_models", description = 'Merge multiple model files')
    parser.add_argument('models', nargs="+")
    parser.add_argument('-o','--output', required=True)

    # merge_dicts
    parser = subparsers.add_parser("merge_dicts", description = 'Merge restraint dictionary cif files')
    parser.add_argument('cifs', nargs="+")
    parser.add_argument('-o','--output', default="merged.cif", help="Output cif file (default: %(default)s)")

    # geom
    parser = subparsers.add_parser("geom", description = 'Calculate geometry and show outliers')
    parser.add_argument('model')
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('--sigma', type=float, default=5,
                        help="sigma cutoff to print outliers (default: %(default).1f)")
    parser.add_argument('--per_atom_score_as_b', action='store_true',
                        help="write model file with per-atom score as B values")
    parser.add_argument("--check_skew", action='store_true', help="(experimental) check bond skew to test magnification")
    parser.add_argument('-n', '--nucleus', action="store_true", help="Use nucleus distances (for neutron)")
    parser.add_argument("--ignore_h", action='store_true', help="ignore hydrogen")
    parser.add_argument("--selection", help="evaluate part of the model")
    parser.add_argument('-o', '--output_prefix', 
                        help="default: taken from input file")

    # adp
    parser = subparsers.add_parser("adp", description = 'ADP analysis')
    parser.add_argument('model')
    parser.add_argument('-o', '--output_prefix',
                        help="default: taken from input file")
    
    # power
    parser = subparsers.add_parser("power", description = 'Show power spectrum')
    parser.add_argument("--map",  nargs="*", action="append")
    parser.add_argument("--halfmaps",  nargs="*", action="append")
    parser.add_argument('--mask', help='Mask file')
    parser.add_argument('-d', '--resolution', type=float)
    parser.add_argument('-o', '--output_prefix', default="power")

    # fcalc
    parser = subparsers.add_parser("fcalc", description = 'Structure factor from model')
    parser.add_argument('--model', required=True)
    parser.add_argument("--no_expand_ncs", action='store_true', help="Do not expand strict NCS in MTRIX or _struct_ncs_oper")
    parser.add_argument("--method", choices=["fft", "direct"], default="fft")
    parser.add_argument("--source", choices=["electron", "xray", "neutron"], default="electron")
    parser.add_argument('--ligand', nargs="*", action="append")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--cell', type=float, nargs=6, metavar=("a", "b", "c", "alpha", "beta", "gamma"),
                        help="Override unit cell")
    parser.add_argument('--auto_box_with_padding', type=float, help="Determine box size from model with specified padding")
    parser.add_argument('--cutoff', type=float, default=1e-5)
    parser.add_argument('--rate', type=float, default=1.5)
    parser.add_argument('--add_dummy_sigma', action='store_true', help="write dummy SIGF")
    parser.add_argument('--as_intensity', action='store_true', help="if you want |F|^2")
    parser.add_argument('--keep_charges',  action='store_true',
                        help="Use scattering factor for charged atoms. Use it with care.")
    parser.add_argument('-d', '--resolution', type=float, required=True)
    parser.add_argument('-o', '--output_prefix')

    # nemap
    parser = subparsers.add_parser("nemap", description = 'Normalized expected map calculation from half maps')
    parser.add_argument("--halfmaps", required=True, nargs=2)
    parser.add_argument('--pixel_size', type=float, help='Override pixel size (A)')
    parser.add_argument("--half1_only", action='store_true', help="Only use half 1 for map calculation (use half 2 only for noise estimation)")
    parser.add_argument('-B', type=float, help="local B value")
    parser.add_argument("--no_fsc_weights", action='store_true',
                        help="Just for debugging purpose: turn off FSC-based weighting")
    parser.add_argument("--sharpening_b", type=float,
                        help="Use B value (negative value for sharpening) instead of standard deviation of the signal")
    parser.add_argument("-d", '--resolution', type=float)
    parser.add_argument('-m', '--mask', help="mask file")
    parser.add_argument('-o', '--output_prefix', default='nemap')
    parser.add_argument("--trim", action='store_true', help="Write trimmed maps")
    parser.add_argument("--trim_mtz", action='store_true', help="Write trimmed mtz")
    parser.add_argument("--local_fourier_weighting_with", type=float, default=0,
                        help="Experimental: give kernel size in A^-1 unit to use local Fourier weighting instead of resolution-dependent weights")

    # blur
    parser = subparsers.add_parser("blur", description = 'Blur data by specified B value')
    parser.add_argument('--hklin', required=True, help="input MTZ file")
    parser.add_argument('-B', type=float, required=True, help="B value for blurring (negative value for sharpening)")
    parser.add_argument('-o', '--output_prefix')

    # mask_from_model
    parser = subparsers.add_parser("mask_from_model", description = 'Make a mask from model')
    parser.add_argument("--map", required=True, help="For unit cell and pixel size reference")
    parser.add_argument("--model", required=True)
    parser.add_argument("--selection")
    parser.add_argument('--radius', type=float, required=True,
                        help='Radius in angstrom')
    parser.add_argument('--soft_edge', type=float, default=0,
                        help='Soft edge (default: %(default).1f)')
    parser.add_argument('-o', '--output', default="mask_from_model.mrc")
    
    # applymask (and normalize within mask)
    parser = subparsers.add_parser("applymask", description = 'Apply mask and optionally normalize map within mask')
    parser.add_argument("--map", required=True)
    parser.add_argument('--mask', required=True, help='Mask file')
    parser.add_argument("--normalize", action='store_true',
                        help="Normalize map values using mean and sd within the mask")
    parser.add_argument("--trim", action='store_true', help="Write trimmed map")
    parser.add_argument('--mask_cutoff', type=float, default=0.5,
                        help="cutoff value for normalization and trimming (default: %(default)s)")
    parser.add_argument('-o', '--output_prefix')

    # map2mtz
    parser = subparsers.add_parser("map2mtz", description = 'FFT map and write an mtz')
    parser.add_argument("--map", required=True)
    parser.add_argument("-d", '--resolution', type=float)
    parser.add_argument('-o', '--output')

    # sm2mm
    parser = subparsers.add_parser("sm2mm", description = 'Small molecule files (cif/hkl/res/ins) to macromolecules (pdb/mmcif/mtz)')
    parser.add_argument('files', nargs='+', help='Cif/ins/res/hkl files')
    parser.add_argument('-o', '--output_prefix')

    # seq
    parser = subparsers.add_parser("seq", description = 'Print/align model sequence')
    parser.add_argument("--model", required=True)
    parser.add_argument('--seq', nargs="*", action="append", help="Sequence file(s)")

    # dnarna
    parser = subparsers.add_parser("dnarna", description = 'DNA to RNA or RNA to DNA model conversion')
    parser.add_argument("model")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--to_dna', action='store_true', help="To DNA")
    group.add_argument('--to_rna', action='store_true', help="To RNA")
    parser.add_argument('--chains', nargs="*", action="append", help="Select chains to convert")
    parser.add_argument('-o', '--output')

# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def symmodel(args):
    if args.chains: args.chains = sum(args.chains, [])
    model_format = fileio.check_model_format(args.model)

    howtoname = dict(dup=gemmi.HowToNameCopiedChain.Dup,
                     short=gemmi.HowToNameCopiedChain.Short,
                     number=gemmi.HowToNameCopiedChain.AddNumber)[args.howtoname]

    if (args.twist, args.rise).count(None) == 1:
        raise SystemExit("ERROR: give both helical parameters --twist and --rise")

    is_helical = args.twist is not None
    st, cif_ref = fileio.read_structure_from_pdb_and_mmcif(args.model)
    st.spacegroup_hm = "P 1"
    map_and_start = None
    if args.map:
        logger.writeln("Reading cell from map")
        map_and_start = fileio.read_ccp4_map(args.map)
        st.cell = map_and_start[0].unit_cell
    elif args.cell:
        st.cell = gemmi.UnitCell(*args.cell)
    elif not st.cell.is_crystal():
        raise SystemExit("Error: Unit cell parameters look wrong. Please use --map or --cell")

    if args.chains:
        logger.writeln("Keep {} chains only".format(" ".join(args.chains)))
        chains = set(args.chains)
        for m in st:
            to_del = [c.name for c in m if c.name not in chains]
            for c in to_del: m.remove_chain(c)
        if st[0].count_atom_sites() == 0:
            raise SystemExit("ERROR: no atoms left. Check --chains option.")

    all_chains = [c.name for c in st[0] if c.name not in st[0]]

    symmetry.update_ncs_from_args(args, st, map_and_start=map_and_start, filter_contacting=args.contacting_only)

    if args.biomt:
        st.assemblies.clear()
        st.raw_remarks = []
        a = model.prepare_assembly("1", all_chains, st.ncs, is_helical=is_helical)
        st.assemblies.append(a)

    if not args.output_prfix:
        args.output_prfix = fileio.splitext(os.path.basename(args.model))[0] + "_asu"

    if args.pdb or args.cif:
        fileio.write_model(st, args.output_prfix, pdb=args.pdb, cif=args.cif, cif_ref=cif_ref)
    else:
        fileio.write_model(st, file_name=args.output_prfix+model_format, cif_ref=cif_ref)

    # Sym expand
    model.expand_ncs(st, howtoname=howtoname)
    st.assemblies.clear()
    args.output_prfix += "_expanded"
    if args.pdb or args.cif:
        fileio.write_model(st, args.output_prfix, pdb=args.pdb, cif=args.cif)
    else:
        fileio.write_model(st, file_name=args.output_prfix+model_format)
# symmodel()

def helical_biomt(args):
    if (args.twist, args.rise).count(None) > 0:
        raise SystemExit("ERROR: give helical parameters --twist and --rise")

    model_format = fileio.check_model_format(args.model)
    howtoname = dict(dup=gemmi.HowToNameCopiedChain.Dup,
                     short=gemmi.HowToNameCopiedChain.Short,
                     number=gemmi.HowToNameCopiedChain.AddNumber)[args.howtoname]

    st, cif_ref = fileio.read_structure_from_pdb_and_mmcif(args.model)
    st.spacegroup_hm = "P 1"
    map_and_start = None
    if args.map:
        logger.writeln("Reading cell from map")
        map_and_start = fileio.read_ccp4_map(args.map)
        st.cell = map_and_start[0].unit_cell
    elif args.cell:
        st.cell = gemmi.UnitCell(*args.cell)
    elif not st.cell.is_crystal():
        raise SystemExit("Error: Unit cell parameters look wrong. Please use --map or --cell")

    all_chains = [c.name for c in st[0] if c.name not in st[0]]

    ncsops = symmetry.ncsops_from_args(args, st.cell, map_and_start=map_and_start, st=st,
                                       helical_min_n=args.start, helical_max_n=args.end)
    #ncsops = [x for x in ncsops if not x.tr.is_identity()] # remove identity

    logger.writeln("")
    logger.writeln("-------------------------------------------------------------")
    logger.writeln("You may need to write following matrices in OneDep interface:")
    for idx, op in enumerate(ncsops):
        logger.writeln("")
        logger.writeln("operator {}".format(idx+1))
        mat = op.tr.mat.tolist()
        vec = op.tr.vec.tolist()
        for i in range(3):
            mstr = ["{:10.6f}".format(mat[i][j]) for j in range(3)]
            logger.writeln("{} {:14.5f}".format(" ".join(mstr), vec[i]))
    logger.writeln("-------------------------------------------------------------")
    logger.writeln("")

    # BIOMT
    st.assemblies.clear()
    st.raw_remarks = []
    a = model.prepare_assembly("1", all_chains, ncsops, is_helical=True)
    st.assemblies.append(a)

    if not args.output_prfix:
        args.output_prfix = fileio.splitext(os.path.basename(args.model))[0] + "_biomt"

    fileio.write_model(st, args.output_prfix, pdb=(model_format == ".pdb"), cif=True, cif_ref=cif_ref)
    logger.writeln("")
    logger.writeln("These {}.* files may be used for deposition (once OneDep implemented reading BIOMT from file..)".format(args.output_prfix))
    logger.writeln("")
    # BIOMT expand
    st.transform_to_assembly("1", howtoname)
    args.output_prfix += "_expanded"
    fileio.write_model(st, file_name=args.output_prfix+model_format)
    logger.writeln(" note that this expanded model file is just for visual inspection, *not* for deposition!")
# helical_biomt()

def symexpand(args):
    if args.chains: args.chains = sum(args.chains, [])
    model_format = fileio.check_model_format(args.model)
    if not args.split:
        howtoname = dict(dup=gemmi.HowToNameCopiedChain.Dup,
                         short=gemmi.HowToNameCopiedChain.Short,
                         number=gemmi.HowToNameCopiedChain.AddNumber)[args.howtoname]

    st = fileio.read_structure(args.model)

    if args.chains:
        logger.writeln("Keep {} chains only".format(" ".join(args.chains)))
        chains = set(args.chains)
        for m in st:
            to_del = [c.name for c in m if c.name not in chains]
            for c in to_del: m.remove_chain(c)

    all_chains = [c.name for c in st[0] if c.name not in st[0]]

    if not args.output_prfix:
        args.output_prfix = fileio.splitext(os.path.basename(args.model))[0]

    if len(st.ncs) > 0:
        symmetry.show_ncs_operators_axis_angle(st.ncs)
        non_given = [op for op in st.ncs if not op.given]
        if len(non_given) > 0:
            if args.split:
                for i, op in enumerate(st.ncs):
                    if op.given: continue
                    st_tmp = st.clone()
                    for m in st_tmp: m.transform_pos_and_adp(op.tr)
                    output_prfix = args.output_prfix + "_ncs_{:02d}".format(i+1)
                    if args.pdb or args.cif:
                        fileio.write_model(st_tmp, output_prfix, pdb=args.pdb, cif=args.cif)
                    else:
                        fileio.write_model(st_tmp, file_name=output_prfix+model_format)
            else:
                st_tmp = st.clone()
                model.expand_ncs(st_tmp, howtoname=howtoname)
                output_prfix = args.output_prfix + "_ncs_expanded"
                if args.pdb or args.cif:
                    fileio.write_model(st_tmp, output_prfix, pdb=args.pdb, cif=args.cif)
                else:
                    fileio.write_model(st_tmp, file_name=output_prfix+model_format)
        else:
            logger.writeln("All operators are already expanded (marked as given). Exiting.")
    else:
        logger.writeln("No NCS operators found. Exiting.")
    
    if len(st.assemblies) > 0: # should we support BIOMT?
        pass
# symexpand()

def h_add(args):
    st = fileio.read_structure(args.model)
    model_format = fileio.check_model_format(args.model)
    
    if not args.output:
        tmp = fileio.splitext(os.path.basename(args.model))[0]
        args.output = tmp + "_h" + model_format
    logger.writeln("Output file: {}".format(args.output))
        
    args.ligand = sum(args.ligand, []) if args.ligand else []
    monlib = restraints.load_monomer_library(st,
                                             monomer_dir=args.monlib,
                                             cif_files=args.ligand)
    model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)
    try:
        restraints.add_hydrogens(st, monlib, args.pos)
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    fileio.write_model(st, file_name=args.output)
# h_add()

def add_op3(args):
    st = fileio.read_structure(args.model)
    model_format = fileio.check_model_format(args.model)
    
    if not args.output:
        tmp = fileio.splitext(os.path.basename(args.model))[0]
        args.output = tmp + "_op3" + model_format
    logger.writeln("Output file: {}".format(args.output))
    
    args.ligand = sum(args.ligand, []) if args.ligand else []
    monlib = restraints.load_monomer_library(st,
                                             monomer_dir=args.monlib,
                                             cif_files=args.ligand)
    model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)

    for chain in st[0]:
        p = chain.get_polymer()
        if not p: continue
        p_type = p.check_polymer_type()
        if p_type not in (gemmi.PolymerType.Dna, gemmi.PolymerType.Rna): continue
        r0 = p[0]
        # TODO: alias
        # TODO: altlocs
        alt = "*"
        if r0.find_atom("OP3", alt): continue
        a_op1 = r0.find_atom("OP1", alt)
        a_op2 = r0.find_atom("OP2", alt)
        a_o5p = r0.find_atom("O5'", alt)
        a_p = r0.find_atom("P", alt)
        if None in (a_op1, a_op2, a_o5p, a_p):
            logger.writeln(f"Error: atoms not found. skipping {chain.name}/{r0}")
            continue
        logger.writeln(f"Adding OP3 to {chain.name}/{r0}")
        a_op3 = r0.add_atom(a_p) # inherit ADP and occupancy
        a_op3.name = "OP3"
        a_op3.element = gemmi.Element("O")
        v1 = a_p.pos - a_op1.pos
        v2 = a_p.pos - a_op2.pos
        v3 = a_p.pos - a_o5p.pos
        v = v1 + v2 + v3
        a_op3.pos = a_p.pos + v / v.length() * 1.517
    
    fileio.write_model(st, file_name=args.output)
# add_op3()

def read_map_and_oversample(map_in=None, mtz_in=None, mtz_labs=None, oversample_pixel=None):
    if mtz_in is not None:
        mtz = fileio.read_mmhkl(mtz_in)
        lab_f, lab_phi = mtz_labs.split(",")
        asu = mtz.get_f_phi(lab_f, lab_phi)
        if oversample_pixel is not None:
            d_min = numpy.min(asu.make_d_array())
            sample_rate = d_min / oversample_pixel
        else:
            sample_rate = 3
        gr = asu.transform_f_phi_to_map(sample_rate=sample_rate)
    elif map_in is not None:
        gr = fileio.read_ccp4_map(map_in)[0]
        if oversample_pixel is not None:
            asu = gemmi.transform_map_to_f_phi(gr).prepare_asu_data()
            d_min = numpy.min(asu.make_d_array())
            sample_rate = d_min / oversample_pixel
            gr = asu.transform_f_phi_to_map(sample_rate=sample_rate)
    else:
        raise SystemExit("Invalid input")
    
    if oversample_pixel is not None:
        logger.writeln("--oversample_pixel= {} is requested.".format(oversample_pixel))
        logger.writeln(" recalculated grid:")
        logger.writeln("  {:4d} {:4d} {:4d}".format(*gr.shape))
        logger.writeln(" spacings:")
        logger.writeln("  {:.6f} {:.6f} {:.6f}".format(*gr.spacing))
        #maps.write_ccp4_map("{}_oversampled.mrc".format(output_prefix), gr)

    return gr
# read_map_and_oversample()    

def map_peaks(args):
    st = fileio.read_structure(args.model)
    gr = read_map_and_oversample(map_in=args.map, mtz_in=args.mtz, mtz_labs=args.mtz_labels,
                                 oversample_pixel=args.oversample_pixel)
    gr_sigma = numpy.std(gr)
    if args.abs_level is not None:
        cutoff = args.abs_level
    else:
        cutoff = args.sigma_level * gr_sigma # assuming mean(gr) = 0
        
    blobs = gemmi.find_blobs_by_flood_fill(gr, cutoff,
                                           min_volume=args.min_volume, min_score=0)
    blobs.extend(gemmi.find_blobs_by_flood_fill(gr, cutoff, negate=True,
                                                min_volume=args.min_volume, min_score=0))
    getpos = dict(peak=lambda x: x.peak_pos,
                  centroid=lambda x: x.centroid)[args.blob_pos]
    st_peaks = model.st_from_positions([getpos(b) for b in blobs])
    st_peaks.cell = st.cell
    st_peaks.ncs = st.ncs
    st_peaks.setup_cell_images()
    logger.writeln("{} peaks detected".format(len(blobs)))
    #st_peaks.write_pdb("peaks.pdb")
    
    # Filter symmetry related
    ns = gemmi.NeighborSearch(st_peaks[0], st_peaks.cell, 5.).populate()
    cs = gemmi.ContactSearch(1.)
    cs.ignore = gemmi.ContactSearch.Ignore.SameAsu
    results = cs.find_contacts(ns)
    del_idxes = set()
    for r in results:
        if r.partner1.residue.seqid.num not in del_idxes:
            del_idxes.add(r.partner2.residue.seqid.num)
    for i in reversed(sorted(del_idxes)):
        del st_peaks[0][0][i]
        del blobs[i]
    #st_peaks.write_pdb("peaks_asu.pdb")
    logger.writeln("{} peaks after removing symmetry equivalents".format(len(blobs)))

    # Assign to nearest atom
    ns = gemmi.NeighborSearch(st[0], st.cell, 10.).populate() # blob is rejected if > 10 A. ok?
    peaks = []
    for b in blobs:
        bpos = getpos(b)
        map_val = gr.interpolate_value(bpos)
        if (args.max_volume is not None and b.volume > args.max_volume) or abs(map_val) < cutoff: continue
        x = ns.find_nearest_atom(bpos)
        if x is None: # this should not happen
            logger.writeln("no nearest atom: value={:.2e} volume= {:.2f} pos= {}".format(map_val, b.volume, bpos))
            continue
        chain = st[0][x.chain_idx]
        res = chain[x.residue_idx]
        atom = res[x.atom_idx]
        im = st.cell.find_nearest_image(atom.pos, bpos, gemmi.Asu.Any)
        mpos = st.cell.find_nearest_pbc_position(atom.pos, bpos, im.sym_idx)
        dist = atom.pos.dist(mpos)
        peaks.append((map_val, b.volume, mpos, dist, chain, res, atom))

    if len(peaks) == 0:
        logger.writeln("No peaks found. Change parameter(s).")
        return
        
    # Print and write coot script
    peaks.sort(reverse=True, key=lambda x:(abs(x[0]), x[1]))
    for_coot = []
    for_df = []
    for i, p in enumerate(peaks):
        map_val, volume, mpos, dist, chain, res, atom = p
        mpos_str = "({: 7.2f},{: 7.2f},{: 7.2f})".format(mpos.x, mpos.y, mpos.z)
        atom_name = atom.name + ("." + atom.altloc if atom.altloc != "\0" else "")
        atom_str = "{}/{}/{}".format(chain.name, res.seqid, atom_name)
        if args.abs_level is None:
            map_val /= gr_sigma
        lab_str = "Peak {:4d} value= {: .2e} volume= {:5.1f} pos= {} closest= {:10s} dist= {:.2f}".format(i+1, map_val, volume, mpos_str, atom_str, dist)
        for_coot.append((lab_str, (mpos.x, mpos.y, mpos.z)))
        for_df.append((map_val, volume, mpos.x, mpos.y, mpos.z, chain.name, str(res.seqid), atom_name, dist))
    df = pandas.DataFrame(for_df, columns=["map_value" if args.abs_level is not None else "sigma_level",
                                           "volume", "x", "y", "z", "chain", "residue", "atom", "dist"])
    logger.writeln(df.to_string())
    with open(args.output_prefix + ".json", "w") as ofs:
        df.to_json(ofs, orient="records", indent=2)
        logger.writeln("saved: {}".format(ofs.name))
    coot_out = args.output_prefix + "_coot.py"
    with open(coot_out, "w") as ofs:
        ofs.write("""\
from __future__ import absolute_import, division, print_function
import gtk
class coot_serval_map_peak_list:
  def __init__(self):
    window = gtk.Window(gtk.WINDOW_TOPLEVEL)
    window.set_title("Map peaks (Servalcat)")
    window.set_default_size(600, 600)
    scrolled_win = gtk.ScrolledWindow()
    scrolled_win.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
    vbox = gtk.VBox(False, 2)
    frame_vbox = gtk.VBox(False, 0)
    frame_vbox.set_border_width(3)
    self.btns = []
    self.data = {}
    self.add_data(frame_vbox)
    scrolled_win.add_with_viewport(frame_vbox)
    vbox.pack_start(scrolled_win, True, True, 0)
    window.add(vbox)
    window.show_all()
    self.toggled(self.btns[0], 0)

  def toggled(self, btn, i):
    if btn.get_active():
      set_rotation_centre(*self.data[i][1])
      add_status_bar_text(self.data[i][0])

  def add_data(self, vbox):
    for i, d in enumerate(self.data):
      self.btns.append(gtk.RadioButton(None if i == 0 else self.btns[0], d[0]))
      vbox.pack_start(self.btns[-1], False, False, 0)
      self.btns[-1].connect('toggled', self.toggled, i)

gui = coot_serval_map_peak_list()
""".format(for_coot))
    logger.writeln("\nRun:")
    logger.writeln("coot --script {}".format(coot_out))
# map_peaks()

def h_density_analysis(args):
    #if args.source != "electron":
    #    raise SystemExit("Only electron source is supported.")
    model_format = fileio.check_model_format(args.model)
    st = fileio.read_structure(args.model)
    if not st[0].has_hydrogen():
        raise SystemExit("No hydrogen in model.")

    if args.output_prefix is None:
        args.output_prefix = fileio.splitext(os.path.basename(args.model))[0] + "_hana"

    gr = read_map_and_oversample(map_in=args.map, mtz_in=args.mtz, mtz_labs=args.mtz_labels,
                                 oversample_pixel=args.oversample_pixel)
                            
    if args.abs_level is not None:
        cutoff = args.abs_level
    else:
        cutoff = args.sigma_level * numpy.std(gr) # assuming mean(gr) = 0

    blobs = gemmi.find_blobs_by_flood_fill(gr, cutoff,
                                           min_volume=args.min_volume, min_score=0)
    getpos = dict(peak=lambda x: x.peak_pos,
                  centroid=lambda x: x.centroid)[args.blob_pos]

    peaks = [getpos(b).tolist() for b in blobs]
    kdtree = scipy.spatial.cKDTree(peaks)
    found = []
    n_hydr = 0
    h_assigned = [0 for _ in range(len(blobs))]
    st2 = st.clone()
    for ic, chain in enumerate(st[0]):
        for ir, res in enumerate(chain):
            for ia, atom in reversed(list(enumerate(res))):
                if not atom.is_hydrogen(): continue
                n_hydr += 1
                dist, idx = kdtree.query(atom.pos.tolist(), k=1, p=2)
                map_val = gr.interpolate_value(getpos(blobs[idx]))
                if dist < args.max_dist and blobs[idx].volume < args.max_volume and map_val > cutoff:
                    found.append((getpos(blobs[idx]), map_val, dist,  blobs[idx].volume, 
                                  chain.name, str(res.seqid), res.name,
                                  atom.name, atom.altloc.replace("\0","")))
                    h_assigned[idx] = 1
                else:
                    del st2[0][ic][ir][ia]

    found.sort(key=lambda x: x[1], reverse=True)
    logger.writeln("")
    logger.writeln("Found hydrogen peaks:")
    logger.writeln("dist map  vol  atom")
    for _, map_val, dist, volume, chain, resi, resn, atom, alt in found:
        logger.writeln("{:.2f} {:.2f} {:.2f} {}/{} {}/{}{}".format(dist, map_val, volume,
                                                                 chain, resn, resi,
                                                                 atom, "."+alt if alt else ""))

    logger.writeln("")
    logger.writeln("Result:")
    logger.writeln(" number of hydrogen in the model  : {}".format(n_hydr))
    logger.writeln(" number of peaks close to hydrogen: {} ({:.1%})".format(len(found), len(found)/n_hydr))
    logger.writeln("")

    st_peaks = model.st_from_positions([getpos(b) for b in blobs],
                                       bs=[gr.interpolate_value(getpos(b)) for b in blobs],
                                       qs=h_assigned)
    fileio.write_model(st_peaks, file_name="{}_peaks.mmcif".format(args.output_prefix))
    logger.writeln(" this file includes peak positions")
    logger.writeln(" occ=1: hydrogen assigned, occ=0: unassigned.")
    logger.writeln(" B: density value at {}".format(args.blob_pos))
    logger.writeln("")
    
    fileio.write_model(st2, file_name="{}_h_with_peak{}".format(args.output_prefix, model_format))
    logger.writeln(" this file is a copy of input model, where hydrogen atoms without peaks are removed.")
# h_density_analysis()

def fix_link(args):
    st = fileio.read_structure(args.model)
    model_format = fileio.check_model_format(args.model)
    
    if not args.output:
        tmp = fileio.splitext(os.path.basename(args.model))[0]
        args.output = tmp + "_fixlink" + model_format
    logger.writeln("Output file: {}".format(args.output))
        
    args.ligand = sum(args.ligand, []) if args.ligand else []
    monlib = restraints.load_monomer_library(st,
                                             monomer_dir=args.monlib,
                                             cif_files=args.ligand)
    model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)
    restraints.find_and_fix_links(st, monlib, bond_margin=args.bond_margin,
                                  metal_margin=args.metal_margin)
    fileio.write_model(st, file_name=args.output)
# fix_link()
    
def merge_models(args):
    logger.writeln("Reading file   1: {}".format(args.models[0]))
    st = fileio.read_structure(args.models[0])
    logger.writeln("                  chains {}".format(" ".join([c.name for c in st[0]])))

    for i, f in enumerate(args.models[1:]):
        logger.writeln("Reading file {:3d}: {}".format(i+2, f))
        st2 = fileio.read_structure(f)
        for c in st2[0]:
            org_id = c.name
            c2 = st[0].add_chain(c, unique_name=True)
            if c.name != c2.name:
                logger.writeln("                  chain {} merged (ID changed to {})".format(c.name, c2.name))
            else:
                logger.writeln("                  chain {} merged".format(c.name))

    fileio.write_model(st, file_name=args.output)
# merge_models()

def merge_dicts(args):
    fileio.merge_ligand_cif(args.cifs, args.output)
# merge_dicts()

def geometry(args):
    if args.ligand: args.ligand = sum(args.ligand, [])
    if not args.output_prefix: args.output_prefix = fileio.splitext(os.path.basename(args.model))[0] + "_geom"
    keywords = []
    if args.keywords or args.keyword_file:
        if args.keywords: keywords = sum(args.keywords, [])
        if args.keyword_file: keywords.extend(l for f in sum(args.keyword_file, []) for l in open(f))
    params = refmac_keywords.parse_keywords(keywords)
    st = fileio.read_structure(args.model)
    if args.ignore_h:
        st.remove_hydrogens()
    try:
        monlib = restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand, 
                                                 stop_for_unknowns=True, params=params)
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)
    restraints.find_and_fix_links(st, monlib)
    try:
        topo, _ = restraints.prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.NoChange,
                                              check_hydrogen=True, params=params)
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))
    
    if args.selection:
        sel = gemmi.Selection(args.selection)
        atom_pos = [-1 for _ in range(st[0].count_atom_sites())]
        n = 0
        for chain in sel.chains(st[0]):
            for res in sel.residues(chain):
                for atom in sel.atoms(res):
                    atom_pos[atom.serial-1] = n
                    n += 1
        logger.writeln("Using selection '{}': {} atoms out of {}".format(args.selection, n, len(atom_pos)))
    else:
        atom_pos = None

    geom = Geom(st, topo, monlib, params=params, atom_pos=atom_pos, use_nucleus=args.nucleus)
    for k in geom.outlier_sigmas: geom.outlier_sigmas[k] = args.sigma
    geom.setup_nonbonded(True)
    ret = geom.show_model_stats()
    
    with open(args.output_prefix + "_summary.json", "w") as ofs:
        ret["summary"].to_json(ofs, indent=2)
        logger.writeln("saved: {}".format(ofs.name))
    with open(args.output_prefix + "_outliers.json", "w") as ofs:
        for k in ret["outliers"]:
            ret["outliers"][k] = ret["outliers"][k].to_dict(orient="records")
        json.dump(ret["outliers"], ofs, indent=2)
        logger.writeln("saved: {}".format(ofs.name))

    if args.check_skew:
        logger.writeln("\nChecking skewness of bond length deviation")
        # better to ignore hydrogen
        tab = geom.geom.reporting.get_bond_outliers(use_nucleus=geom.use_nucleus, min_z=0)
        for a in "atom1", "atom2":
            tab[a] = [str(geom.lookup[x]) for x in tab[a]]
        df = pandas.DataFrame(tab)
        df["dev"] = df["value"] - df["ideal"]
        df = df.reindex(df.dev.abs().sort_values(ascending=False).index)
        logger.writeln("Bond length deviations:")
        logger.writeln(df.to_string(max_rows=20))
        q1, q2, q3 = numpy.percentile(df["dev"], [25, 50, 75])
        sk2 = (q1 + q3 - 2 * q2) / (q3 - q1)
        logger.writeln("bond_dev_median= {:.6f}".format(q2))
        logger.writeln("bond_dev_skew=   {:.4f}".format(df["dev"].skew()))
        logger.writeln("bond_dev_sk2=    {:.4f}".format(sk2))
        with open(args.output_prefix + "_bond_dev.html", "w") as ofs:
            ofs.write("""\
<html>
<head>
 <meta charset="utf-8" />
 <script src="https://cdn.plot.ly/plotly-2.20.0.min.js" charset="utf-8"></script>
</head>
<body>
 <div id="hist"></div>
 <script>
  var trace = {
   x: %s,
   type: 'histogram'
  };
  var layout = {
   title: "median: %.4f, sk2: %.4f",
   xaxis: {title: "bond distance - ideal"},
   yaxis: {title: "count"},
   shapes: [{
    type: 'line',
    yref: 'paper',
    x0: 0, y0: 0,
    x1: 0, y1: 1}]
  };
  target = document.getElementById('hist');
  Plotly.newPlot(target, [trace], layout);
 </script>
</body>
</html>
""" % (str(list(df.dev)), q2, sk2))
            logger.writeln("check histogram: {}".format(ofs.name))

    # Note that this modifies st
    if args.per_atom_score_as_b:
        model_format = fileio.check_model_format(args.model)
        peratom = geom.geom.reporting.per_atom_score(len(geom.atoms), geom.use_nucleus, "mean")
        for i, score in enumerate(peratom["total"]):
            geom.atoms[i].b_iso = score
        fileio.write_model(st, file_name="{}_per_atom_score{}".format(args.output_prefix, model_format))
# geometry()

def adp_stats(args):
    if not args.output_prefix: args.output_prefix = fileio.splitext(os.path.basename(args.model))[0] + "_adp"
    st = fileio.read_structure(args.model)
    model.adp_analysis(st)
    b_all = [cra.atom.b_iso for cra in st[0].all() if cra.atom.occ > 0]
    
    # bin width from Freedman–Diaconis rule
    qs = numpy.quantile(b_all, [0, 0.25, 0.75, 1])
    bin_h = 2 * (qs[2] - qs[1]) / len(b_all)**(1/3.)
    
    # for plotly
    traces = []
    traces.append("x: [%s], type: 'histogram', name: 'All', xbins: {size: %f}"
                  % (",".join("%.2f"%x for x in b_all), bin_h))
    if len(st[0]) > 1:
        b_chain = {}
        for c in st[0]:
            b_chain.setdefault(c.name, []).extend(a.b_iso for r in c for a in r if a.occ > 0)
        for c in b_chain:
            bs = ",".join("%.2f" % x for x in b_chain[c])
            traces.append("x: [%s], type: 'histogram', name: 'Chain %s'" % (bs, c))
    with open(args.output_prefix + "_hist.html", "w") as ofs:
        ofs.write("""\
<html>
<head>
 <meta charset="utf-8" />
 <script src="https://cdn.plot.ly/plotly-2.20.0.min.js" charset="utf-8"></script>
</head>
<body>
 <div id="hist"></div>
 <script>
""")
        for i, t in enumerate(traces):
            ofs.write("var trace%d = {%s};\n" % (i+1, t))
        ofs.write("""\
  var layout = {
   title: "isotropic B histogram",
   xaxis: {title: "B"},
   yaxis: {title: "count"},
   barmode: "stack"
  };
  target = document.getElementById('hist');
  Plotly.newPlot(target, [%s], layout);
 </script>
</body>
</html>
""" % (",".join("trace%d" % (i+1) for i in range(len(traces)))))
        logger.writeln("check histogram: {}".format(ofs.name))
# adp_stats()

def show_power(args):
    maps_in = []
    if args.map:
        print(args.map)
        print(sum(args.map, []))
        maps_in = [(f,) for f in sum(args.map, [])]
        
    if args.halfmaps:
        args.halfmaps = sum(args.halfmaps, [])
        if len(args.halfmaps)%2 != 0:
            raise RuntimeError("Number of half maps is not even.")
        maps_in.extend([(args.halfmaps[2*i],args.halfmaps[2*i+1]) for i in range(len(args.halfmaps)//2)])
        
    if args.mask:
        mask = fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None

    hkldata = None
    labs = []
    for mapin in maps_in: # TODO rewrite in faster way
        ms = [fileio.read_ccp4_map(f) for f in mapin]
        d_min = args.resolution
        if d_min is None:
            d_min = maps.nyquist_resolution(ms[0][0])
            logger.writeln("WARNING: --resolution is not specified. Using Nyquist resolution: {:.2f}".format(d_min))
        tmp = maps.mask_and_fft_maps(ms, d_min, mask)
        labs.append("F{:02d}".format(len(labs)+1))
        tmp.df.rename(columns=dict(FP=labs[-1]), inplace=True)
        if hkldata is None:
            hkldata = tmp
        else:
            if hkldata.cell.parameters != tmp.cell.parameters: raise RuntimeError("Different unit cell!")
            hkldata.merge(tmp.df[["H","K","L",labs[-1]]])

    if not labs:
        raise SystemExit("No map files given. Exiting.")
            
    hkldata.setup_relion_binning()

    ofs = open(args.output_prefix+".log", "w")
    ofs.write("Input:\n")
    for i in range(len(maps_in)):
        ofs.write("{} from {}\n".format(labs[i], " ".join(maps_in[i])))
    ofs.write("\n")
    
    ofs.write("""$TABLE: Power spectrum :
$GRAPHS
: log10(Mn(|F|^2)) :A:1,{}:
$$
1/resol^2 n d_max d_min {}
$$
$$
""".format(",".join([str(i+5) for i in range(len(labs))]), " ".join(labs)))
    print(hkldata.df)
    abssqr = dict((lab, numpy.abs(hkldata.df[lab].to_numpy())**2) for lab in labs)
    for i_bin, idxes in hkldata.binned():
        bin_d_min = hkldata.binned_df.d_min[i_bin]
        bin_d_max = hkldata.binned_df.d_max[i_bin]
        ofs.write("{:.4f} {:7d} {:7.3f} {:7.3f}".format(1/bin_d_min**2, len(idxes), bin_d_max, bin_d_min,))
        for lab in labs:
            pwr = numpy.log10(numpy.average(abssqr[lab][idxes]))
            ofs.write(" {:.4e}".format(pwr))
        ofs.write("\n")
    ofs.write("$$\n")
    ofs.close()
# show_power()

def fcalc(args):
    if (args.auto_box_with_padding, args.cell).count(None) == 0:
        raise SystemExit("Error: you cannot specify both --auto_box_with_padding and --cell")
    
    if args.ligand: args.ligand = sum(args.ligand, [])
    if not args.output_prefix: args.output_prefix = "{}_fcalc_{}".format(fileio.splitext(os.path.basename(args.model))[0], args.source)

    st = fileio.read_structure(args.model)
    if not args.keep_charges:
        model.remove_charge([st])
    model.check_atomsf([st], args.source)
    if not args.no_expand_ncs:
        model.expand_ncs(st)    

    if args.cell is not None:
        st.cell = gemmi.UnitCell(*args.cell)
    elif args.auto_box_with_padding is not None:
        st.cell = model.box_from_model(st[0], args.auto_box_with_padding)
        st.spacegroup_hm = "P 1"
        logger.writeln("Box size from the model with padding of {}: {}".format(args.auto_box_with_padding, st.cell.parameters))
        
    if not st.cell.is_crystal():
        raise SystemExit("ERROR: No unit cell information. Give --cell or --auto_box_with_padding.")

    if args.source=="electron" and st[0].has_hydrogen():
        monlib = restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand, 
                                                 stop_for_unknowns=False)
    else:
        monlib = None

    if args.method == "fft":
        fc_asu = model.calc_fc_fft(st, args.resolution, cutoff=args.cutoff, rate=args.rate,
                                   mott_bethe=args.source=="electron",
                                   monlib=monlib, source=args.source)
    else:
        fc_asu = model.calc_fc_direct(st, args.resolution, source=args.source,
                                      mott_bethe=args.source=="electron", monlib=monlib)

    hkldata = hkl.hkldata_from_asu_data(fc_asu, "FC")
    if args.as_intensity:
        hkldata.df["IC"] = numpy.abs(hkldata.df.FC)**2
        labout = ["IC"]
        if args.add_dummy_sigma:
            hkldata.df["SIGIC"] = 1.
            labout.append("SIGIC")
    else:
        labout = ["FC"]
        if args.add_dummy_sigma:
            hkldata.df["SIGFC"] = 1.
            labout.append("SIGFC")

    hkldata.write_mtz(args.output_prefix+".mtz", labout, types=dict(IC="J", SIGIC="Q", SIGFC="Q"))
# fcalc()

def nemap(args):
    from servalcat.spa import fofc

    if (args.trim or args.trim_mtz) and args.mask is None:
        raise SystemExit("\nError: You need to give --mask as you requested --trim or --trim_mtz.\n")
    
    if args.mask:
        mask = fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None

    halfmaps = fileio.read_halfmaps(args.halfmaps, pixel_size=args.pixel_size)
    if args.resolution is None:
        args.resolution = maps.nyquist_resolution(halfmaps[0][0])
        logger.writeln("WARNING: --resolution is not specified. Using Nyquist resolution: {:.2f}".format(args.resolution))

    d_min = args.resolution
    if args.local_fourier_weighting_with > 0:
        d_min = 1 / (args.local_fourier_weighting_with + 1 / d_min)
        logger.writeln("adjusting d_min= {:.2f} for local correlation".format(d_min))
    hkldata = maps.mask_and_fft_maps(halfmaps, d_min, mask)

    if args.local_fourier_weighting_with > 0:
        asu1 = hkldata.as_asu_data("F_map1")
        asu2 = hkldata.as_asu_data("F_map2")
        size = asu1.get_size_for_hkl(sample_rate=3)
        logger.writeln("using grid {}".format(size))
        gr1 = asu1.get_f_phi_on_grid(size)
        gr2 = asu2.get_f_phi_on_grid(size)
        kernel = ext.hard_sphere_kernel_recgrid(size, asu1.unit_cell, args.local_fourier_weighting_with)
        cc = maps.local_cc(gr1, gr2, kernel.array.real, method="simple")
        cc.array[cc.array < 0] = 0 # negative cc cannot be used anyway
        cc.array[:] = 2 * cc.array.real / (1 + cc.array.real) # to full map cc
        hkldata.df["cc"] = numpy.real(cc.get_value_by_hkl(hkldata.miller_array()))
        grf = type(gr1)((gr1.array + gr2.array) / 2, gr1.unit_cell, gr1.spacegroup)
        var_f = maps.local_var(grf, kernel.array.real, method="simple")
        hkldata.df["var_f"] = numpy.real(var_f.get_value_by_hkl(hkldata.miller_array()))
        if args.B is not None:
            k2_l = numpy.exp(-args.B / hkldata.d_spacings()**2 / 2)
            hkldata.df.cc = k2_l * hkldata.df.cc / (1 + (k2_l - 1) * hkldata.df.cc)
        hkldata.df["FWT"] = hkldata.df.FP * numpy.sqrt(hkldata.df.cc / hkldata.df.var_f)
        hkldata.df["kernel"] = numpy.real(kernel.get_value_by_hkl(hkldata.miller_array()))
        hkldata.write_mtz(args.output_prefix+"_cc.mtz", ["cc", "kernel"])
        hkldata = hkldata.copy(d_min=args.resolution)
        map_labs = ["FWT"]
    else:
        hkldata.setup_relion_binning()
        maps.calc_noise_var_from_halfmaps(hkldata)
        map_labs = fofc.calc_maps(hkldata, B=args.B, has_halfmaps=True, half1_only=args.half1_only,
                                  no_fsc_weights=args.no_fsc_weights, sharpening_b=args.sharpening_b)
    fofc.write_files(hkldata, map_labs, grid_start=halfmaps[0][1], stats_str=None,
                     mask=mask, output_prefix=args.output_prefix,
                     trim_map=args.trim, trim_mtz=args.trim_mtz)
# nemap()

def blur(args):
    if args.output_prefix is None:
        args.output_prefix = fileio.splitext(os.path.basename(args.hklin))[0]
    
    if fileio.is_mmhkl_file(args.hklin):
        mtz = fileio.read_mmhkl(args.hklin)
        hkl.blur_mtz(mtz, args.B)
        suffix = ("_blur" if args.B > 0 else "_sharpen") + "_{:.2f}.mtz".format(abs(args.B))
        mtz.write_to_file(args.output_prefix+suffix)
        logger.writeln("Written: {}".format(args.output_prefix+suffix))
    else:
        raise SystemExit("ERROR: Unsupported file type: {}".format(args.hklin))
# blur()

def mask_from_model(args):
    st = fileio.read_structure(args.model) # TODO option to (or not to) expand NCS
    if args.selection:
        gemmi.Selection(args.selection).remove_not_selected(st)
    gr, grid_start, _ = fileio.read_ccp4_map(args.map)
    mask = maps.mask_from_model(st, args.radius, soft_edge=args.soft_edge, grid=gr)
    maps.write_ccp4_map(args.output, mask, grid_start=grid_start)
# mask_from_model()

def applymask(args):
    if args.output_prefix is None:
        args.output_prefix = fileio.splitext(os.path.basename(args.map))[0] + "_masked"

    grid, grid_start, _ = fileio.read_ccp4_map(args.map)
    mask = fileio.read_ccp4_map(args.mask)[0]
    logger.writeln("Applying mask")
    logger.writeln(" mask min: {:.3f} max: {:.3f}".format(numpy.min(mask), numpy.max(mask)))
    grid.array[:] *= mask.array

    if args.normalize:
        masked = grid.array[mask.array>args.mask_cutoff]
        masked_mean = numpy.average(masked)
        masked_std = numpy.std(masked)
        logger.writeln("Normalizing map values within mask")
        logger.writeln(" masked volume: {} mean: {:.3e} sd: {:.3e}".format(len(masked), masked_mean, masked_std))
        grid.array[:] = (grid.array - masked_mean) / masked_std

    maps.write_ccp4_map(args.output_prefix+".mrc", grid,
                        grid_start=grid_start,
                        mask_for_extent=mask.array if args.trim else None,
                        mask_threshold=args.mask_cutoff)
# applymask()

def map2mtz(args):
    if args.output is None:
        args.output = fileio.splitext(os.path.basename(args.map))[0] + "_fft.mtz"
    grid, grid_start, grid_shape = fileio.read_ccp4_map(args.map)
    if args.resolution is None:
        args.resolution = maps.nyquist_resolution(grid)
        logger.writeln("WARNING: --resolution is not specified. Using Nyquist resolution: {:.2f}".format(args.resolution))

    if grid_start != (0,0,0) or grid.shape != tuple(grid_shape):
        # If only subregion of whole grid in map, unit cell needs to be re-defined.
        if grid.shape != tuple(grid_shape):
            new_abc = [grid.unit_cell.parameters[i] * grid_shape[i] / grid.shape[i] for i in range(3)]
            cell = gemmi.UnitCell(*new_abc, *grid.unit_cell.parameters[3:])
            logger.writeln("Changing unit cell to {}".format(cell.parameters))
        else:
            cell = grid.unit_cell
        grid = gemmi.FloatGrid(grid.get_subarray(grid_start, grid_shape),
                               cell, grid.spacegroup)
    
    f_grid = gemmi.transform_map_to_f_phi(grid)
    asudata = f_grid.prepare_asu_data(dmin=args.resolution, with_000=True)
    hkldata = hkl.hkldata_from_asu_data(asudata, "F")
    if grid_start != (0,0,0):
        shifts = grid.get_position(*grid_start)
        hkldata.translate("F", shifts)
        logger.writeln("Applying phase shift with translation {}".format(shifts.tolist()))
    hkldata.write_mtz(args.output, ["F"])
# map2mtz()

def sm2mm(args):
    if args.output_prefix is None:
        args.output_prefix = fileio.splitext(args.files[0])[0]
    st, mtz = fileio.read_small_molecule_files(args.files)
    if st is not None:
        fileio.write_model(st, prefix=args.output_prefix, pdb=True, cif=True)
    if mtz is not None:
        mtz_out = args.output_prefix + ".mtz"
        logger.writeln("Writing MTZ file: {}".format(mtz_out))
        mtz.write_to_file(mtz_out)
# sm2mm()

def seq(args):
    wrap_width = 100
    seqs = []
    if args.seq:
        args.seq = sum(args.seq, [])
        for sf in args.seq:
            seqs.extend(fileio.read_sequence_file(sf))
        
    st = fileio.read_structure(args.model) # TODO option to (or not to) expand NCS
    model.setup_entities(st, clear=True, force_subchain_names=True, overwrite_entity_type=True)    
    for chain in st[0]:
        p = chain.get_polymer()
        if not p: continue
        p_type = p.check_polymer_type()
        if p_type in (gemmi.PolymerType.SaccharideD, gemmi.PolymerType.SaccharideL): continue
        p_seq = gemmi.one_letter_code(p.extract_sequence())
        results = []
        for name, seq in seqs:
            # what if DnaRnaHybrid?
            kind = {gemmi.PolymerType.Dna: gemmi.ResidueKind.DNA,
                    gemmi.PolymerType.Rna: gemmi.ResidueKind.RNA}.get(p_type, gemmi.ResidueKind.AA)
            s = [gemmi.expand_one_letter(x, kind) for x in seq]
            if None in s: continue
            results.append([name, gemmi.align_sequence_to_polymer(s, p, p_type), seq])

        if results:
            logger.writeln("Chain: {}".format(chain.name))
            logger.writeln(" polymer type: {}".format(str(p_type).replace("PolymerType.", "")))
            name, al, s1 = max(results, key=lambda x: x[1].score)
            logger.writeln(" match: {}".format(name))
            logger.writeln(" score: {}".format(al.score))
            p1, p2 = al.add_gaps(s1, 1), al.add_gaps(p_seq, 2)
            unkseq = [x.start() for x in re.finditer("\-", p1)]
            mismatches = [x.start() for x in re.finditer("\.", al.match_string)]
            if mismatches or unkseq:
                idxes = {x.start(): i for i, x in enumerate(re.finditer("[^-]", p2))}
                seqnums = [str(x.seqid) for x in p]
                if mismatches:
                    logger.write(" mismatches: ")
                    logger.writeln(", ".join("{}({}>{})".format(seqnums[idxes[i]], p1[i], p2[i]) for i in mismatches))
                if unkseq:
                    logger.write(" unknown sequence: ")
                    logger.writeln(", ".join("{}({})".format(seqnums[idxes[i]], p2[i]) for i in unkseq))

            logger.writeln("")
            for i in range(0, len(p1), wrap_width):
                logger.writeln(" seq.  {}".format(p1[i:i+wrap_width]))
                logger.writeln("       {}".format(al.match_string[i:i+wrap_width]))
                logger.writeln(" model {}\n".format(p2[i:i+wrap_width]))
        else:
            logger.writeln("> Chain: {}".format(chain.name))
            logger.writeln(gemmi.one_letter_code(p.extract_sequence()))
            logger.writeln("")
# seq()

def dnarna(args):
    import scipy.spatial.transform
    rna_res = {"A":"DA", "G":"DG", "C":"DC", "U":"DT"}
    dna_res = {"DA":"A", "DG":"G", "DC":"C", "DT":"U"}
    if args.chains: args.chains = sum(args.chains, [])
    model_format = fileio.check_model_format(args.model)
    if not args.output:
        args.output = fileio.splitext(os.path.basename(args.model))[0] + "_conv" + model_format
    st = fileio.read_structure(args.model)
    if st[0].has_hydrogen():
        logger.writeln("Hydrogen atoms are detected. I cannot take care of them, so I will remove them.")
        st.remove_hydrogens()
    for chain in st[0]:
        if args.chains and chain.name not in args.chains:
            continue
        for res in chain:
            alt = "*" # XXX
            if res.name in rna_res and args.to_dna:
                logger.writeln(f"Changing {chain.name}/{res.seqid} {res.name} to DNA")
                res.name = rna_res[res.name]
                res.remove_atom("O2'", alt)
                if res.name == "DT":
                    C4 = res.find_atom("C4", alt)
                    C5 = res.find_atom("C5", alt)
                    C6 = res.find_atom("C6", alt)
                    v1 = C5.pos - C4.pos
                    v2 = C5.pos - C6.pos
                    v = v1 + v2
                    res.add_atom(C5)
                    res[-1].name = "C7"
                    res[-1].pos = C5.pos + v / v.length() * 1.5
            elif res.name in dna_res and args.to_rna:
                logger.writeln(f"Changing {chain.name}/{res.seqid} {res.name} to RNA")
                res.name = dna_res[res.name]
                C1p = numpy.array(res.find_atom("C1'", alt).pos.tolist())
                C2p = numpy.array(res.find_atom("C2'", alt).pos.tolist())
                C3p = numpy.array(res.find_atom("C3'", alt).pos.tolist())
                rotvec = C2p - C3p
                rotvec /= numpy.linalg.norm(rotvec)
                r = scipy.spatial.transform.Rotation.from_rotvec(-rotvec * 120,
                                                                 degrees=True)
                rotated = r.apply(C1p - C2p)
                rotated *= 1.411 / numpy.linalg.norm(rotated)
                res.add_atom(res.find_atom("O3'", alt))
                res[-1].name = "O2'"
                res[-1].pos.fromlist(C2p + rotated)
                if res.name == "U":
                    res.remove_atom("C7", alt)
    fileio.write_model(st, file_name=args.output)
# dnarna()

def show(args):
    for filename in args.files:
        ext = fileio.splitext(filename)[1]
        if ext in (".mrc", ".ccp4", ".map"):
            fileio.read_ccp4_map(filename)
            logger.writeln("\n")
# show()

def json2csv(args):
    if not args.output_prefix:
        args.output_prefix = fileio.splitext(os.path.basename(args.json))[0]
        
    df = pandas.read_json(args.json)
    df.to_csv(args.output_prefix+".csv", index=False)
    logger.writeln("Output: {}".format(args.output_prefix+".csv"))
# json2csv()

def main(args):
    comms = dict(show=show,
                 json2csv=json2csv,
                 symmodel=symmodel,
                 helical_biomt=helical_biomt,
                 expand=symexpand,
                 h_add=h_add,
                 add_op3=add_op3,
                 map_peaks=map_peaks,
                 h_density=h_density_analysis,
                 fix_link=fix_link,
                 merge_models=merge_models,
                 merge_dicts=merge_dicts,
                 geom=geometry,
                 adp=adp_stats,
                 power=show_power,
                 fcalc=fcalc,
                 nemap=nemap,
                 blur=blur,
                 mask_from_model=mask_from_model,
                 applymask=applymask,
                 map2mtz=map2mtz,
                 sm2mm=sm2mm,
                 seq=seq,
                 dnarna=dnarna)
    
    com = args.subcommand
    f = comms.get(com)
    if f:
        return f(args)
    else:
        raise SystemExit("Unknown subcommand: {}".format(com))
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
import numpy

def cgsolve_rm(A, v, M, gamma=0., ncycl=2000, toler=1.e-4):
    gamma_save = None
    step_flag = False
    gamma_flag = False
    conver_flag = False
    #f = numpy.zeros(len(v))
    dv = numpy.zeros(len(v))
    dv_save = numpy.zeros(len(v))
    
    # preconditioning
    A = M.T.dot(A).dot(M)
    #print("precond_A=")
    #print(A.toarray())
    v = M.T.dot(v)

    vnorm = numpy.sqrt(numpy.dot(v, v))
    test_lim = toler * vnorm
    max_gamma_cyc = 500
    step = 0.05

    for gamma_cyc in range(max_gamma_cyc):
        if gamma_cyc != 0: gamma += step

        logger.writeln("Trying gamma equal {:.4e}".format(gamma))
        r = v - (A.dot(dv) + gamma * dv)
        rho = [numpy.dot(r, r)]
        if rho[0] < toler:
            break
        
        exit_flag = False
        for itr in range(ncycl):
            if itr == 0:
                p = r
                beta = 0.
            else:
                beta = rho[-1] / rho[-2]
                p = r + beta * p

            f = A.dot(p) + gamma * p
            alpha = rho[-1] / numpy.dot(p, f)
            dv += alpha * p
            if itr%20 == 0:
                r = v - (A.dot(dv) + gamma * dv)
            else:
                r = r - alpha * f

            rho.append(numpy.dot(r, r))
            #print("rho=", rho)
            if numpy.sqrt(rho[-1]) > 2 * numpy.sqrt(rho[-2]):
                logger.writeln("Not converging with gamma equal {:.4e}".format(gamma))
                step *= 1.05
                break

            if numpy.sqrt(rho[-1]) < test_lim:
                if not gamma_flag:
                    logger.writeln("Convergence reached with no gamma cycles")
                    exit_flag = True
                    break # goto 120
                elif conver_flag:
                    logger.writeln("Convergence reached with gamma equal {:.4e}".format(gamma))
                    step *= 1.01
                    exit_flag = True
                    break # goto 120
                else:
                    conver_flag = True
                    step_flag = True
                    gamma_save = gamma
                    dv_save = numpy.copy(dv)
                    gamma = max(0, gamma - step/5.)
                    step = max(step/1.1, 0.0001)
                    logger.writeln("Gamma decreased to {:.4e}".format(gamma))
                    exit_flag = True
                    break
        # end of inner loop
        if exit_flag: break
        
        gamma_flag = True
        if not conver_flag:
            dv[:] = 0.
        else:
            dv = numpy.copy(dv_save)
            gamma = gamma_save
            logger.writeln("Back to gamma equal {:.4e}".format(gamma))


    # postconditioning
    dv = M.dot(dv)
    return dv, gamma
# cgsolve_rm()



# TODO shift map first using mask!
"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import time
from servalcat.utils import logger
from servalcat import utils
from servalcat.spa import shift_maps
import argparse

def add_arguments(parser):
    parser.description = 'Fo-Fc map calculation based on model and data errors'
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--halfmaps", nargs=2)
    group.add_argument("--map", help="Use only if you really do not have half maps.")
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('--model', required=True,
                        help='Input atomic model file')
    parser.add_argument("-d", '--resolution', type=float, required=True)
    parser.add_argument('-m', '--mask', help="mask file")
    parser.add_argument('-r', '--mask_radius', type=float, help="mask radius (not used if --mask is given)")
    parser.add_argument('--no_check_mask_with_model', action='store_true', 
                        help='Disable mask test using model')
    parser.add_argument("-B", type=float, help="Estimated blurring")
    parser.add_argument("--half1_only", action='store_true', help="Only use half 1 for map calculation (use half 2 only for noise estimation)")
    parser.add_argument("--normalized_map", action='store_true',
                        help="Write normalized map in the masked region. Now this is on by default.")
    parser.add_argument("--no_fsc_weights", action='store_true',
                        help="Just for debugging purpose: turn off FSC-based weighting")
    parser.add_argument("--sharpening_b", type=float,
                        help="Use B value (negative value for sharpening) instead of standard deviation of the signal")
    parser.add_argument("--trim", action='store_true',
                        help="Write trimmed maps")
    parser.add_argument("--trim_mtz", action='store_true',
                        help="Write trimmed mtz")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument("--omit_proton", action='store_true',
                        help="Omit proton from model in map calculation")
    parser.add_argument("--omit_h_electron", action='store_true',
                        help="Omit hydrogen electrons from model in map calculation")
    parser.add_argument("--source", choices=["electron", "xray", "neutron"], default="electron")
    parser.add_argument('-o','--output_prefix', default="diffmap",
                        help='output file name prefix (default: %(default)s)')
    parser.add_argument('--keep_charges',  action='store_true',
                        help="Use scattering factor for charged atoms. Use it with care.")
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_D_and_S(hkldata, has_halfmaps=True, half1_only=False):#fo_asu, fc_asu, varn, bins, bin_idxes):
    bdf = hkldata.binned_df
    bdf["D"] = 0.
    bdf["S"] = 0.
    stats_str = """$TABLE: Statistics :
$GRAPHS
: log(Mn(|F|^2)) and variances :A:1,6,7,8,13,14:
: FSC :A:1,9,10,11:
: weights :A:1,12,15,16:
: map weights :A:1,17:
$$
1/resol^2 bin n d_max d_min log(var(Fo)) log(var(Fc)) log(var(DFc)) FSC.model FSC.full sqrt(FSC.full) D log(var_U,T) log(var_noise) wFo wFc wFo.sharpen
$$
$$
"""
    tmpl = "{:.4f} {:3d} {:7d} {:7.3f} {:7.3f} {:.4e} {:.4e} {:4e} {: .4f}   {: .4f} {: .4f} {: .4e} {:.4e} {:.4e} {:.4f} {:.4f} {:.4e}\n"

    var_noise = None
    FP = hkldata.df.FP.to_numpy()
    if half1_only:
        FP = hkldata.df.F_map1.to_numpy()
        var_noise = hkldata.binned_df.var_noise * 2
    elif has_halfmaps:
        var_noise = hkldata.binned_df.var_noise
        
    for i_bin, idxes in hkldata.binned():
        bin_d_min = hkldata.binned_df.d_min[i_bin]
        bin_d_max = hkldata.binned_df.d_max[i_bin]
        Fo = FP[idxes]
        Fc = hkldata.df.FC.to_numpy()[idxes]
        fsc = numpy.real(numpy.corrcoef(Fo, Fc)[1,0])
        bdf.loc[i_bin, "D"] = numpy.sum(numpy.real(Fo * numpy.conj(Fc)))/numpy.sum(numpy.abs(Fc)**2)
        if has_halfmaps:
            varn = var_noise[i_bin]
            fsc_full = hkldata.binned_df.FSCfull[i_bin]
            S = max(0, numpy.average(numpy.abs(Fo-bdf.D[i_bin]*Fc)**2)-varn)
            bdf.loc[i_bin, "S"] = S
            w = S/(S+varn)
            if fsc_full < 0: # this should be fixed actually. needs smoothing to zero.
                w_sharpen = 0
            else:
                w_sharpen = w / numpy.sqrt(fsc_full) / numpy.std(Fo)
        else:
            varn = fsc_full = 0
            w = 1
            w_sharpen = 1

        with numpy.errstate(divide="ignore", invalid="ignore"):
            stats_str += tmpl.format(1/bin_d_min**2, i_bin, Fo.size, bin_d_max, bin_d_min,
                                     numpy.log(numpy.average(numpy.abs(Fo)**2)),
                                     numpy.log(numpy.average(numpy.abs(Fc)**2)),
                                     numpy.log(bdf.D[i_bin]**2*numpy.average(numpy.abs(Fc)**2)),
                                     fsc, fsc_full, numpy.sqrt(fsc_full), bdf.D[i_bin],
                                     numpy.log(bdf.S[i_bin]), numpy.log(varn),
                                     w, 1-w, w_sharpen)
    return stats_str
# calc_D_and_S()

#import line_profiler
#profile = line_profiler.LineProfiler()
#import atexit
#atexit.register(profile.print_stats)
#@profile
def calc_maps(hkldata, B=None, has_halfmaps=True, half1_only=False, no_fsc_weights=False, sharpening_b=None):
    has_fc = "FC" in hkldata.df

    if not has_fc:
        labs = ["FWT"]
        if B is not None: labs.append("FWT_b0")
    elif has_halfmaps:
        labs = ["Fupdate", "DELFWT", "FWT", "DELFWT_noscale", "Fupdate_noscale"]
        if B is not None: labs.extend(["Fupdate_b0", "DELFWT_b0", "FWT_b0"])
    else:
        labs = ["DELFWT"]
        
    tmp = {}
    for l in labs:
        tmp[l] = numpy.zeros(len(hkldata.df.index), numpy.complex128)

    logger.writeln("Calculating maps..")
    logger.write(" sharpening method: ")
    if sharpening_b is None:
        logger.writeln("1/sqrt(FSC * Mn(Fo)) for Fo and 1/sigma_U,T for Fo-Fc")
    else:
        logger.writeln("1/exp(-B*s^2/4) with B= {:.2f}".format(sharpening_b))

    time_t = time.time()

    if half1_only:
        FP = hkldata.df.F_map1.to_numpy()
    else:
        FP = hkldata.df.FP.to_numpy()

    s2 = 1./hkldata.d_spacings().to_numpy()**2

    fsc_became_negative = False
        
    for i_bin, idxes in hkldata.binned():
        if has_halfmaps:
            fsc = hkldata.binned_df.FSCfull[i_bin] # FSCfull
            if half1_only:
                varn = hkldata.binned_df.var_noise[i_bin] * 2
                fsc = fsc/(2-fsc) # to FSChalf
            else:
                varn = hkldata.binned_df.var_noise[i_bin]
        else:
            fsc, varn = 1., 0.
            
        w_nomodel = 1. if no_fsc_weights else fsc
        Fo = FP[idxes]
        sig_fo = numpy.std(Fo)
        s2_bin = s2[idxes]

        if has_fc:
            Fc = hkldata.df.FC.to_numpy()[idxes]
            D = hkldata.binned_df.D[i_bin]
            S = hkldata.binned_df.S[i_bin] # variance of unexplained signal
            w = 1. if no_fsc_weights or not has_halfmaps else S/(S+varn)
            delfwt = w * (Fo-D*Fc)
            fup = 2 * w * Fo + (1 - 2*w) * D*Fc # <F> + delfwt
            if has_halfmaps: # no point making this map when half maps not given
                tmp["DELFWT_noscale"][idxes] = delfwt
                tmp["Fupdate_noscale"][idxes] = fup

        if not fsc_became_negative and fsc <= 0:
            logger.writeln(" WARNING: cutting resolution at {:.2f} A because fsc < 0".format(hkldata.binned_df.d_max[i_bin]))
            fsc_became_negative = True
        if fsc_became_negative:
            continue

        if sharpening_b is None:
            k = sig_fo * numpy.sqrt(fsc)
            k_fofc = numpy.sqrt(S) if has_fc and S > 0 else 1. # to avoid zero-division. if S=0 then w=0.
        else:
            k = k_fofc = numpy.exp(-sharpening_b*s2_bin/4)
            
        lab_suf = "" if B is None else "_b0"
        if has_halfmaps:
            tmp["FWT"+lab_suf][idxes] = w_nomodel / k * Fo
            if has_fc:
                tmp["DELFWT"+lab_suf][idxes] = delfwt / k_fofc
                tmp["Fupdate"+lab_suf][idxes] = fup / k
        elif has_fc:
            tmp["DELFWT"+lab_suf][idxes] = delfwt

        if B is not None and has_halfmaps: # local B based map
            k_l = numpy.exp(-B*s2_bin/4.)
            k2_l = numpy.exp(-B*s2_bin/2.)
            fsc_l = k2_l*fsc/(1+(k2_l-1)*fsc)
            w_nomodel = 1. if no_fsc_weights else fsc_l
            tmp["FWT"][idxes] = Fo*w_nomodel/k/k_l
            if has_fc:
                S_l = S * k2_l
                w = 1. if no_fsc_weights or not has_halfmaps else S_l/(S_l+varn)            
                delfwt = (Fo-D*Fc)*w/k_fofc/k_l
                fup = (w*Fo+(1.-w)*D*Fc)/k/k_l
                logger.writeln("{:4d} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e}".format(i_bin,
                                                                                      numpy.average(k),
                                                                                      numpy.average(sig_fo),
                                                                                      numpy.average(fsc_l),
                                                                                      numpy.average(k_l),
                                                                                      numpy.average(abs(fup)),
                                                                                      numpy.average(abs(delfwt))))
                tmp["DELFWT"][idxes] = delfwt
                tmp["Fupdate"][idxes] = fup

    for l in labs:
        hkldata.df[l] = tmp[l]

    logger.writeln(" finished in {:.3f} sec.".format(time.time()-time_t))
    return labs
# calc_maps()

def dump_to_mtz(hkldata, map_labs, mtz_out):
    extra_labs = list(filter(lambda x: x in hkldata.df, ["FP", "FC"]))
    map_labs = map_labs + extra_labs
    hkldata.write_mtz(mtz_out, map_labs)
# dump_to_mtz()

def calc_fofc(st, d_min, maps, mask=None, monlib=None, B=None, half1_only=False,
              no_fsc_weights=False, sharpening_b=None, omit_proton=False, omit_h_electron=False,
              source="electron"):
    if no_fsc_weights:
        logger.writeln("WARNING: --no_fsc_weights is requested.")
    if sharpening_b is not None:
        logger.writeln("WARNING: --sharpening_b={} is given".format(sharpening_b))
    
    hkldata = utils.maps.mask_and_fft_maps(maps, d_min, mask)
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, d_min - 1e-6, monlib=monlib, source=source,
                                               miller_array=hkldata.miller_array())
    if mask is not None:
        fc_map = hkldata.fft_map("FC", grid_size=mask.shape)
        fc_map.array[:] *= mask
        hkldata.df["FC"] = gemmi.transform_map_to_f_phi(fc_map).get_value_by_hkl(hkldata.miller_array())
        
    hkldata.setup_relion_binning()

    has_halfmaps = (len(maps) == 2)
    if has_halfmaps:
        utils.maps.calc_noise_var_from_halfmaps(hkldata)

    stats_str = calc_D_and_S(hkldata, has_halfmaps=has_halfmaps, half1_only=half1_only)

    if omit_proton or omit_h_electron:
        hkldata.df["FC"] = utils.model.calc_fc_fft(st, d_min - 1e-6, monlib=monlib, source=source,
                                                   omit_proton=omit_proton, omit_h_electron=omit_h_electron,
                                                   miller_array=hkldata.miller_array())
    
    map_labs = calc_maps(hkldata, B=B, has_halfmaps=has_halfmaps, half1_only=half1_only,
                         no_fsc_weights=no_fsc_weights, sharpening_b=sharpening_b)
    return hkldata, map_labs, stats_str
# calc_fofc()

def write_files(hkldata, map_labs, grid_start, stats_str,
                mask=None, output_prefix="diffmap", trim_map=False, trim_mtz=False,
                normalize_map=True, omit_h_electron=False):
    # this function may modify the overall scale of FWT/DELFWT.

    if mask is not None and (trim_map or trim_mtz):
        new_cell, new_shape, new_grid_start, shifts = shift_maps.determine_shape_and_shift(mask=gemmi.FloatGrid(mask.array,
                                                                                                                hkldata.cell,
                                                                                                                hkldata.sg),
                                                                                           grid_start=grid_start,
                                                                                           padding=5,
                                                                                           mask_cutoff=0.5,
                                                                                           noncentered=True,
                                                                                           noncubic=True,
                                                                                           json_out=None)
    else:
        new_cell, new_shape, new_grid_start, shifts = None, None, None, None

    if trim_map:
        grid_start_for_map = new_grid_start
        shape_for_map = new_shape
    else:
        grid_start_for_map = grid_start
        shape_for_map = None
        
    if normalize_map and mask is not None:
        cutoff = 0.5
        if "DELFWT" in hkldata.df:
            logger.writeln("Normalized Fo-Fc map requested.")
            delfwt_map = hkldata.fft_map("DELFWT", grid_size=mask.shape)
            masked = delfwt_map.array[mask.array>cutoff]
            logger.writeln("   Whole volume: {} voxels".format(delfwt_map.point_count))
            logger.writeln("  Masked volume: {} voxels (>{})".format(masked.size, cutoff))
            global_mean = numpy.average(delfwt_map)
            global_std = numpy.std(delfwt_map)
            logger.writeln("    Global mean: {:.3e}".format(global_mean))
            logger.writeln("     Global std: {:.3e}".format(global_std))
            masked_mean = numpy.average(masked)
            masked_std = numpy.std(masked)
            logger.writeln("    Masked mean: {:.3e}".format(masked_mean))
            logger.writeln("     Masked std: {:.3e}".format(masked_std))
            #logger.writeln(" If you want to scale manually: {}".format())
            scaled = (delfwt_map - masked_mean)/masked_std
            hkldata.df["DELFWT"] /= masked_std # it would work if masked_mean~0
            if omit_h_electron:
                scaled *= -1
                filename = "{}_normalized_fofc_flipsign.mrc".format(output_prefix)
            else:
                filename = "{}_normalized_fofc.mrc".format(output_prefix)
            logger.writeln("  Writing {}".format(filename))
            utils.maps.write_ccp4_map(filename, scaled, cell=hkldata.cell,
                                      grid_start=grid_start_for_map, grid_shape=shape_for_map)

        # Write Fo map as well
        if "FWT" in hkldata.df:
            fwt_map = hkldata.fft_map("FWT", grid_size=mask.shape)
            masked = fwt_map.array[mask.array>cutoff]
            masked_mean = numpy.average(masked)
            masked_std = numpy.std(masked)
            scaled = (fwt_map - masked_mean)/masked_std # does not make much sense for Fo map though
            hkldata.df["FWT"] /= masked_std # it would work if masked_mean~0
            filename = "{}_normalized_fo.mrc".format(output_prefix)
            logger.writeln("  Writing {}".format(filename))
            utils.maps.write_ccp4_map(filename, scaled, cell=hkldata.cell,
                                      grid_start=grid_start_for_map, grid_shape=shape_for_map)

    if trim_mtz and shifts is not None:
        hkldata2 = utils.hkl.HklData(new_cell, hkldata.sg, df=None)
        d_min = hkldata.d_min_max()[0]
        for lab in map_labs + ["FP", "FC"]:
            if lab not in hkldata.df: continue
            gr = hkldata.fft_map(lab, grid_size=mask.shape)
            gr = gemmi.FloatGrid(gr.get_subarray(new_grid_start, new_shape),
                                 new_cell, hkldata.sg)
            if hkldata2.df is None:
                ad = gemmi.transform_map_to_f_phi(gr).prepare_asu_data(dmin=d_min)
                hkldata2.merge_asu_data(ad, lab)
            else:
                hkldata2.df[lab] = gemmi.transform_map_to_f_phi(gr).get_value_by_hkl(hkldata2.miller_array())
            hkldata2.translate(lab, -shifts)
        hkldata = hkldata2
        
    dump_to_mtz(hkldata, map_labs, "{}.mtz".format(output_prefix))
    if stats_str: open("{}_Fstats.log".format(output_prefix), "w").write(stats_str)
# write_files()

def write_coot_script(py_out, model_file, mtz_file, contour_fo=1.2, contour_fofc=3.0, ncs_ops=None):
    with open(py_out, "w") as ofs:
        ofs.write('imol = read_pdb("{}")\n'.format(model_file)) # TODO safer
        ofs.write('imol_fo = make_and_draw_map("{}", "FWT", "PHWT", "", 0, 0)\n'.format(mtz_file))
        ofs.write('imol_fofc = make_and_draw_map("{}", "DELFWT", "PHDELWT", "", 0, 1)\n'.format(mtz_file))
        if contour_fo is not None:
            ofs.write('set_contour_level_absolute(imol_fo, {:.1f})\n'.format(contour_fo))
        if contour_fofc is not None:
            ofs.write('set_contour_level_absolute(imol_fofc, {:.1f})\n'.format(contour_fofc))
        if ncs_ops is not None:
            for op in ncs_ops:
                if op.given: continue
                c, resid = utils.symmetry.find_center_of_origin(op.tr.mat, op.tr.vec)
                if resid.length() > 1e-6: continue # coot does not support translation..
                v = [y for x in op.tr.mat.tolist() for y in x] + c.tolist()
                ofs.write("add_molecular_symmetry(imol, {})\n".format(",".join(str(x) for x in v)))
# write_coot_script()

def main(args):
    if not args.halfmaps and not args.map:
        raise SystemExit("Error: give --halfmaps or --map")

    if not args.halfmaps and args.B is not None:
        raise SystemExit("Error: -B only works with half maps")

    if args.half1_only:
        if not args.halfmaps:
            raise SystemExit("--half1_only requires half maps")
        logger.error("--half1_only specified. Half map 2 is used only for noise estimation")

    if args.normalized_map:
        logger.writeln("DeprecationWarning: --normalized_map is now on by default. This option will be removed in the future.")
        
    if not args.halfmaps:
        logger.error("Warning: using --halfmaps is strongly recommended!")

    st = utils.fileio.read_structure(args.model)
    if not args.keep_charges:
        utils.model.remove_charge([st])
    ncs_org = gemmi.NcsOpList(st.ncs)
    utils.model.expand_ncs(st)

    if (args.omit_proton or args.omit_h_electron) and not st[0].has_hydrogen():
        raise SystemExit("ERROR! --omit_proton/--omit_h_electron requested, but no hydrogen atoms were found.")

    if args.halfmaps:
        maps = utils.fileio.read_halfmaps(args.halfmaps, pixel_size=args.pixel_size)
        has_halfmaps = True
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]
        has_halfmaps = False

    grid_start = maps[0][1]
    g = maps[0][0]
    st.spacegroup_hm = "P1"
    st.cell = g.unit_cell

    if st[0].has_hydrogen():
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib)
    else:
        monlib = None

    if args.mask:
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
        if not args.no_check_mask_with_model:
            if not utils.maps.test_mask_with_model(mask, st):
                raise SystemExit("\nError: Model is out of mask.\n"
                                 "Please check your --model and --mask. You can disable this test with --no_check_mask_with_model.")
    elif args.mask_radius:
        mask = utils.maps.mask_from_model(st, args.mask_radius, grid=g)
        utils.maps.write_ccp4_map("mask_from_model.ccp4", mask)
    else:
        mask = None
        logger.writeln("Warning: Mask is needed for map normalization. Use --mask or --mask_radius if you want normalized map.")

    hkldata, map_labs, stats_str = calc_fofc(st, args.resolution, maps, mask=mask, monlib=monlib, B=args.B,
                                             half1_only=args.half1_only, no_fsc_weights=args.no_fsc_weights,
                                             sharpening_b=args.sharpening_b, omit_proton=args.omit_proton,
                                             omit_h_electron=args.omit_h_electron,
                                             source=args.source)
    write_files(hkldata, map_labs, grid_start, stats_str,
                mask=mask, output_prefix=args.output_prefix,
                trim_map=args.trim, trim_mtz=args.trim_mtz, omit_h_electron=args.omit_h_electron)

    py_out = "{}_coot.py".format(args.output_prefix)
    write_coot_script(py_out, model_file=args.model,
                      mtz_file=args.output_prefix+".mtz",
                      contour_fo=None if mask is None else 1.2,
                      contour_fofc=None if mask is None else 3.0,
                      ncs_ops=ncs_org)
    logger.writeln("\nOpen model and diffmap.mtz with COOT:")
    logger.writeln("coot --script " + py_out)
    if mask is not None:
        logger.writeln("\nWant to list Fo-Fc map peaks? Try:")
        logger.writeln("servalcat util map_peaks --map {}_normalized_fofc.mrc --model {} --abs_level 4.0".format(args.output_prefix, args.model))
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)
